from typing import Dict, Optional
from enum import Enum

from sklearn.base import BaseEstimator

from regmmd.models import (
    GaussianLoc,
    GaussianScale,
    Gaussian,
    Beta,
    BetaA,
    BetaB,
    Binomial,
    Gamma,
    GammaRate,
    GammaShape,
    Poisson,
)
from regmmd.models.base_model import EstimationModel
from regmmd.optimizers import _sgd_estimation
from regmmd.utils import MMDResult


class DefinedModels(Enum):
    GAUSSIAN_LOC = GaussianLoc
    GAUSSIAN_SCALE = GaussianScale
    GAUSSIAN = Gaussian
    BETA = Beta
    BETA_A = BetaA
    BETA_B = BetaB
    BINOMIAL = Binomial
    GAMMA = Gamma
    GAMMA_RATE = GammaRate
    GAMMA_SHAPE = GammaShape
    POISSON = Poisson


class MMDEstimator(BaseEstimator):
    """Estimator using the Maximum Mean Discrepancy criterion.

    Maximum Mean Discrepancy (MMD) is a kernel-based statistical test used to
    compare two probability distributions. This estimator fits a parametric model
    by minimizing the MMD between the empirical distribution of the observed data
    and the model distribution.

    Depending on the type of model provided, the estimator will use either an
    exact gradient descent procedure (for :class:`GaussianLoc`) or a stochastic
    gradient descent approach for general models.

    Parameters
    ----------
    model : EstimationModel
        The parametric estimation model to be fitted, provided as an instance of
        ``EstimationModel``. This model defines the distributional form assumed
        for the data and exposes an ``_init_params`` method used to initialise
        parameters before optimisation.

    par_v : float, optional
        Initial value for the variable parameters of the model. If ``None``, it
        will be initialised automatically by the model's ``_init_params`` method
        when :meth:`fit` is called.

    par_c : float, optional
        Initial value for the constant parameters of the model. If
        ``None``, it will be initialised automatically by the model's
        ``_init_params`` method when :meth:`fit` is called.

    kernel : str, default="Gaussian"
        The kernel function used to compute the MMD. Currently supports
        ``"Gaussian"``, ``"Laplace"`` or ``"Cauchy"``.

    bandwidth : str or float, default="auto"
        The bandwidth of the kernel. If set to ``"auto"``, the bandwidth is
        selected automatically using a heuristic method such as the median
        heuristic.

    solver : dict, optional
        A dictionary specifying the solver parameters for the optimisation
        procedure. Expected keys are:

        - ``"burnin"`` (*int*): Number of burn-in steps before recording results.
        - ``"n_step"`` (*int*): Total number of optimisation steps.
        - ``"stepsize"`` (*float*): Learning rate for gradient updates.
        - ``"epsilon"`` (*float*): Convergence tolerance or regularisation term.

        If ``None``, solver settings must be provided before calling :meth:`fit`.

    random_state : int, optional
        random seed to be passed to the model and any sampler used in the SGD optimizers.

    Attributes
    ----------
    par_v : float
        The variable parameter, updated with the optimised value after fitting.
    par_c : float
        The constant parameter, not updated with the optimised value after fitting.

    Notes
    -----
    - For :class:`GaussianLoc` models, an exact gradient descent routine
      (``_gd_gaussian_loc_exact_estimation``) can be used, when the kernel is ``"Gaussian"`` as wellduring fitting.
    - For all other models, a stochastic gradient descent routine
      (``_sgd_estimation``) is applied instead.
    """

    def __init__(
        self,
        model: DefinedModels | EstimationModel,
        par_v: float = None,
        par_c: float = None,
        kernel: str = "Gaussian",
        bandwidth: str = "auto",
        solver: Optional[Dict] = None,
        random_state: Optional[int] = None,
    ):
        if isinstance(model, str):
            try:
                self.model = DefinedModels[model.upper().replace("-", "_")].value(
                    par_c=par_c, par_v=par_v, random_state=random_state
                )
            except KeyError:
                raise ValueError("model string is not defined by the package.")
        elif isinstance(model, EstimationModel):
            self.model = model
        else:
            raise TypeError("Expected either string or EstimationModel!")

        self.par_v = par_v
        self.par_c = par_c
        self.kernel = kernel
        self.bandwidth = bandwidth
        self.solver = solver

    def fit(self, X) -> MMDResult:
        """Fit the MMD estimation model according to the given training data.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Training input samples.

        Returns
        -------
        res : MMDResult
            A dictionary containing the results of the optimization process, including
            the estimated parameters and the optimization trajectory.
        """
        if self.par_v is None or self.par_c is None:
            pars = self.model._init_params(X)
            self.par_v = pars[0]
            self.par_c = pars[1]

        res = self.model._exact_fit(
            X=X,
            par_v=self.par_v,
            par_c=self.par_c,
            solver=self.solver,
            kernel=self.kernel,
            bandwidth=self.bandwidth,
        )
        if res is None:
            res = _sgd_estimation(
                X=X,
                par_v=self.par_v,
                par_c=self.par_c,
                model=self.model,
                kernel=self.kernel,
                burn_in=self.solver["burnin"],
                n_step=self.solver["n_step"],
                stepsize=self.solver["stepsize"],
                bandwidth=self.bandwidth,
                epsilon=self.solver["epsilon"],
            )
        return res



import numpy as np
from scipy.spatial import distance_matrix
from numpy.typing import NDArray


def K1d_dist(u: NDArray, kernel: str, bandwidth: float = 1) -> NDArray:
    """1-Dimensional Kernel evaluation function

    Parameters
    ----------
    u : ndarray of shape (n_samples,)
        points to be evaluated.

    kernel : string choices=("Gaussian", "Laplace", "Cauchy")
        Which kernel to be used.

    bandwidth: float
        parameter that is used in the kernel
    """
    u = u / bandwidth
    if kernel == "Gaussian":
        return np.exp(-(u**2))
    elif kernel == "Laplace":
        return np.exp(-np.abs(u))
    elif kernel == "Cauchy":
        return 1 / (2 + u**2)
    else:
        raise ValueError


def K1d(x: NDArray, y: NDArray, kernel: str, bandwidth: float = 1) -> NDArray:
    """1-Dimensional Kernel difference evaluation function (K(x-y))

    Parameters
    ----------
    x : ndarray of shape (n_samples,)
        points to be evaluated.

    y : ndarray of shape (n_samples,)
        points to be evaluated.

    kernel : string choices=("Gaussian", "Laplace", "Cauchy")
        Which kernel to be used.

    bandwidth: float
        parameter that is used in the kernel
    """
    u = x[:, None] - y
    return K1d_dist(u, kernel, bandwidth)


def Kmd_dist(u: NDArray, kernel: str, bandwidth: float = 1) -> NDArray:
    """m-Dimensional Kernel evaluation function

    Parameters
    ----------
    u : ndarray of shape (n_samples, m_features)
        points to be evaluated.

    kernel : string choices=("Gaussian", "Laplace", "Cauchy")
        Which kernel to be used.

    bandwidth: float
        parameter that is used in the kernel
    """
    u = u / bandwidth
    u = np.linalg.norm(u, axis=1, ord=2)
    if kernel == "Gaussian":
        return np.exp(-(u**2))
    elif kernel == "Laplace":
        return np.exp(-u)
    elif kernel == "Cauchy":
        return 1 / (2 + u**2)
    else:
        raise ValueError


def Kmd(x: NDArray, y: NDArray, kernel: str, bandwidth: float = 1) -> NDArray:
    """Multi-Dimensional Kernel evaluation function

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_features)
        points to be evaluated.

    y : ndarray of shape (n_samples, n_features)
        points to be evaluated.

    kernel : string choices=("Gaussian", "Laplace", "Cauchy")
        Which kernel to be used.

    bandwidth: float
        parameter that is used in the kernel
    """
    # Compute pairwise Euclidean distances
    u = distance_matrix(x, y, p=2) / bandwidth
    if kernel == "Gaussian":
        return np.exp(-(u**2))
    elif kernel == "Laplace":
        return np.exp(-u)
    elif kernel == "Cauchy":
        return 1 / (2 + u**2)
    else:
        raise ValueError



import numpy as np
from scipy.special import factorial
from regmmd.models.base_model import EstimationModel


class PoissonBase(EstimationModel):
    def __init__(self, lam: float = None, random_state=None):
        self.lam = lam

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=random_state)

    def log_prob(self, x: int):
        if self.lam is None:
            raise ValueError(
                "Lam needs to be defined" + "to be able to calculate the log_prob"
            )

        log_Z = -np.log(factorial(x))
        log_exp = -self.lam
        log_p = x * np.log(self.lam)
        return log_Z + log_exp + log_p

    def sample_n(self, n: int):
        if self.lam is None:
            raise ValueError(
                "Lam needs to be defined" + "to be able to calculate the sample"
            )

        return self.rng.poisson(lam=self.lam, size=(n,))

    def _init_params(self, X):
        if self.lam is None:
            self.lam = 1 / np.median(X)
        return self._get_params()

    def _project_params(self, par_v):
        par_v = max(1e-6, par_v)
        return par_v

    def score(self, x: int):
        if self.lam is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )

        return -1 + x / self.lam

    def update(self, par_v):
        self.lam = par_v

    def _get_params(self):
        par_v = self.lam
        par_c = None
        return par_v, par_c


class Poisson(PoissonBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(lam=par_v, random_state=random_state)



import numpy as np
from numpy.typing import NDArray

from regmmd.utils import MMDResult

from regmmd.optimizers._common import _median_heuristic


def _gd_gaussian_loc_exact_estimation(
    X: NDArray,
    par_v: float,
    par_c: float,
    burn_in: int = 500,
    n_step: int = 1000,
    stepsize: float = 1.0,
    bandwidth: float = 1.0,
    epsilon: float = 1e-4,
) -> MMDResult:
    """Estimate the location parameter of a Gaussian model with Gaussian kernel
    using exact MMD gradient descent.

    Computes the exact MMD gradient analytically for a Gaussian location model
    with known scale parameter ``par_c`` and a Gaussian Kernel, avoiding Monte
    Carlo sampling. Uses AdaGrad updates with a burn-in phase followed by
    Polyak-Ruppert averaging.

    Parameters
    ----------
    X : np.array, shape (n_samples,)
        Univariate observed data.

    par_v : float
        Initial value of the location parameter (mean) to be estimated.

    par_c : float
        Known scale parameter (standard deviation) of the Gaussian model.

    burn_in : int, default=500
        Number of burn-in iterations during which parameter iterates are not
        averaged.

    n_step : int, default=1000
        Number of averaging iterations following the burn-in phase.

    stepsize : float, default=1.0
        Initial step size for the AdaGrad update.

    bandwidth : float or str, default=1.0
        Bandwidth parameter for the Gaussian kernel. If ``"auto"``, the
        bandwidth is selected using the median heuristic.

    epsilon : float, default=1e-4
        Initial accumulated squared gradient norm, used to stabilize the
        AdaGrad step size at the start of optimization.

    Returns
    -------
    res : MMDResult
        Dictionary containing:

        - ``par_v_init`` : initial location parameter.
        - ``par_c_init`` : initial scale parameter.
        - ``stepsize`` : step size used.
        - ``bandwidth`` : bandwidth used (resolved if ``"auto"``).
        - ``estimator`` : Polyak-Ruppert average of location parameter iterates.
        - ``trajectory`` : parameter trajectory of shape ``(burn_in + n_step + 1,)``.
    """
    if bandwidth == "auto":
        bandwidth = _median_heuristic(X)

    norm_grad = epsilon

    res = {
        "par_v_init": np.copy(par_v),
        "par_c_init": np.copy(par_c),
        "stepsize": stepsize,
        "bandwidth": bandwidth,
        "convergence": 1,
    }

    trajectory = np.zeros(shape=(burn_in + n_step + 1,))
    trajectory[0] = par_v

    Z = -4 / np.sqrt(1 + 2 * (par_c**2) / (bandwidth**2))
    denom = 2 * (par_c**2) + bandwidth**2
    for i in range(burn_in):
        diff = X - par_v
        grad = Z * np.mean(diff * np.exp(-np.square(diff) / denom))
        norm_grad += grad**2
        par_v -= stepsize * grad / np.sqrt(norm_grad)
        trajectory[i + 1] = par_v

    par_mean = par_v
    Z = -4 / np.sqrt(1 + 2 * (par_c**2) / (bandwidth**2))
    for i in range(n_step):
        diff = X - par_v
        grad = Z * np.mean(diff * np.exp(-np.square(diff) / denom))
        norm_grad += grad**2
        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_mean = (par_mean * (i + 1) + par_v) / (i + 2)
        trajectory[i + burn_in + 1] = par_mean

    res["estimator"] = par_mean
    res["trajectory"] = trajectory

    return res



import numpy as np
import pytest

from regmmd.models.regression.linear_gaussian import LinearGaussian, LinearGaussianLoc
from regmmd.models.regression.logistic import Logistic
from regmmd.models.regression.gamma import GammaRegressionLoc, GammaRegression
from regmmd.models.regression.poisson import PoissonRegression

from regmmd.models import __all_regression__

RNG = np.random.default_rng(42)
X = RNG.normal(size=(30, 2))
y_cont = RNG.normal(size=(30,))
y_binary = RNG.integers(0, 2, size=(30,)).astype(float)
y_pos = np.abs(y_cont) + 0.1


# --- LinearGaussian ---


def test_linear_gaussian_predict():
    beta = np.array([0.5, -0.3])
    phi = 1.0
    model = LinearGaussian(par_v=np.concatenate([beta, [phi]]))
    preds = model.predict(X)
    assert np.allclose(preds, X @ beta)


def test_linear_gaussian_sample_n():
    model = LinearGaussian(par_v=np.array([0.5, -0.3, 1.0]), random_state=0)
    mu = model.predict(X)
    samples = model.sample_n(30, mu)
    assert samples.shape == (30,)


def test_linear_gaussian_log_prob():
    model = LinearGaussian(par_v=np.array([0.5, -0.3, 1.0]))
    result = model.log_prob(X, y_cont)
    assert isinstance(result, float)


def test_linear_gaussian_score_shape():
    # par_v = [beta0, beta1, phi], so score has 3 columns
    model = LinearGaussian(par_v=np.array([0.5, -0.3, 1.0]))
    score = model.score(X, y_cont)
    assert score.shape == (30, 3)


def test_linear_gaussian_updates():
    # par_v = [beta0, beta1, phi], so score has 3 columns
    model = LinearGaussian(par_v=np.array([0.5, -0.3, 1.0]))
    beta_before = model.beta
    phi_before = model.phi

    model.update(par_v=np.array([1.5, -0.5, 2.0]))
    beta_after = model.beta
    phi_after = model.phi

    assert np.all(beta_before == np.array([0.5, -0.3]))
    assert np.all(beta_after == np.array([1.5, -0.5]))
    assert phi_before == 1.0
    assert phi_after == 2.0


def test_linear_gaussian_project_params():
    # par_v = [beta0, beta1, phi], so score has 3 columns
    model = LinearGaussian(par_v=np.array([0.5, -0.3, 1.0]))
    par_v = model._project_params(np.array([0.5, -0.3, 0.0]))

    assert par_v[-1] > 0
    assert np.all(par_v[:-1] == np.array([0.5, -0.3]))


# --- LinearGaussianLoc ---


def test_linear_gaussian_loc_predict():
    beta = np.array([0.5, -0.3])
    model = LinearGaussianLoc(par_v=beta, par_c=1.0)
    preds = model.predict(X)
    assert np.allclose(preds, X @ beta)


def test_linear_gaussian_loc_score_shape():
    model = LinearGaussianLoc(par_v=np.array([0.5, -0.3]), par_c=1.0)
    score = model.score(X, y_cont)
    assert score.shape == (30, 2)


def test_linear_gaussian_loc_update():
    model = LinearGaussianLoc(par_v=np.zeros(2), par_c=1.0)
    new_beta = np.array([1.0, 2.0])
    model.update(new_beta)
    assert np.allclose(model.beta, new_beta)


# --- Logistic ---


def test_logistic_predict_range():
    beta = np.array([0.5, -0.3])
    model = Logistic(par_v=beta)
    preds = model.predict(X)
    assert preds.shape == (30,)
    assert np.all((preds > 0) & (preds < 1))


def test_logistic_sample_n():
    model = Logistic(par_v=np.array([0.5, -0.3]), random_state=0)
    mu = model.predict(X)
    samples = model.sample_n(30, mu)
    assert samples.shape == (30,)
    assert set(np.unique(samples)).issubset({0, 1})


def test_logistic_log_prob():
    model = Logistic(par_v=np.array([0.5, -0.3]))
    result = model.log_prob(X, y_binary)
    assert isinstance(result, float | np.floating)


def test_logistic_score_shape():
    model = Logistic(par_v=np.array([0.5, -0.3]))
    score = model.score(X, y_binary)
    assert score.shape == (30, 2)


def test_logistic_update():
    beta = np.array([0.1, 0.2])
    model = Logistic(par_v=beta)
    new_param = np.array([1.0, 2.0])
    model.update(new_param)
    assert np.allclose(model.beta, new_param)


def test_logistic_inits_params():
    model = Logistic(par_v=None)
    par_v, par_c = model._init_params(X=X, y=y_binary)
    assert par_v.shape == (2,)
    assert par_c is None


# --- GammaRegression ---


def test_gamma_reg_predict_positive():
    beta = np.array([0.1, 0.2])
    shape = np.array([2.0])
    model = GammaRegression(par_v=np.concatenate((beta, shape)), par_c=None)
    preds = model.predict(X)
    assert preds.shape == (30,)
    assert np.all(preds > 0)


def test_gamma_reg_sample_n():
    beta = np.array([0.1, 0.2])
    shape = np.array([2.0])
    model = GammaRegression(
        par_v=np.concatenate((beta, shape)), par_c=None, random_state=0
    )
    mu = model.predict(X)
    samples = model.sample_n(30, mu)
    assert samples.shape == (30,)
    assert np.all(samples > 0)


def test_gamma_reg_score_shape():
    beta = np.array([0.1, 0.2])
    shape = np.array([2.0])
    model = GammaRegression(par_v=np.concatenate((beta, shape)), par_c=None)
    score = model.score(X, y_pos)
    assert score.shape == (30, 3)


def test_gamma_update():
    beta = np.array([0.1, 0.2])
    shape = 1.0
    model = GammaRegression(par_v=np.concatenate((beta, np.array([shape]))))
    new_param = np.array([1.0, 2.0, 3.0])
    model.update(new_param)
    assert np.allclose(model.beta, new_param[:-1])
    assert model.shape == new_param[-1]


def test_gamma_inits_params():
    model = GammaRegression(par_v=None)
    par_v, par_c = model._init_params(X=X, y=y_pos)
    assert par_v.shape == (3,)
    assert par_c is None


# --- GammaRegressionLoc ---


def test_gamma_reg_loc_predict_positive():
    beta = np.array([0.1, 0.2])
    model = GammaRegressionLoc(par_v=beta, par_c=2.0)
    preds = model.predict(X)
    assert preds.shape == (30,)
    assert np.all(preds > 0)


def test_gamma_reg_loc_sample_n():
    model = GammaRegressionLoc(par_v=np.array([0.1, 0.2]), par_c=2.0, random_state=0)
    mu = model.predict(X)
    samples = model.sample_n(30, mu)
    assert samples.shape == (30,)
    assert np.all(samples > 0)


def test_gamma_reg_loc_score_shape():
    model = GammaRegressionLoc(par_v=np.array([0.1, 0.2]), par_c=2.0)
    score = model.score(X, y_pos)
    assert score.shape == (30, 2)


def test_gamma_loc_update():
    beta = np.array([0.1, 0.2])
    model = GammaRegressionLoc(par_v=beta)
    new_beta = np.array([1.0, 2.0])
    model.update(new_beta)
    assert np.allclose(model.beta, new_beta)


def test_gamma_loc_inits_params():
    model = GammaRegressionLoc(par_v=None)
    par_v, par_c = model._init_params(X=X, y=y_pos)
    assert par_v.shape == (2,)
    assert isinstance(par_c, float)


# --- PoissonRegression ---


def test_poisson_reg_predict_positive():
    beta = np.array([0.1, 0.2])
    model = PoissonRegression(par_v=beta)
    preds = model.predict(X)
    assert preds.shape == (30,)
    assert np.all(preds > 0)


def test_poisson_reg_sample_n():
    beta = np.array([0.1, 0.2])
    model = PoissonRegression(par_v=beta, par_c=None, random_state=0)
    mu = model.predict(X)
    samples = model.sample_n(30, mu)
    assert samples.shape == (30,)
    assert np.all(samples == samples.astype(int))


def test_poisson_log_prob():
    model = PoissonRegression(par_v=np.array([0.5, -0.3]))
    result = model.log_prob(X, y_pos)
    assert isinstance(result, float)


def test_poisson_reg_score_shape():
    beta = np.array([0.1, 0.2])
    model = PoissonRegression(par_v=beta, par_c=None)
    score = model.score(X, y_pos)
    assert score.shape == (30, 2)


def test_poisson_update():
    beta = np.array([0.1, 0.2])
    model = PoissonRegression(par_v=beta)
    new_beta = np.array([1.0, 2.0])
    model.update(new_beta)
    assert np.allclose(model.beta, new_beta)


def test_poisson_inits_params():
    model = PoissonRegression(par_v=None)
    par_v, par_c = model._init_params(X=X, y=y_pos)
    assert par_v.shape == (2,)
    assert par_c is None


# --- Parametrized test ---


@pytest.mark.parametrize("model", __all_regression__)
def test_models_no_par_raises(model):
    model = eval(model)(par_v=None)
    x = np.array([[2, 1], [5, 8], [3, 4]], dtype=float)
    y = np.array([2, 5, 8], dtype=float)
    with pytest.raises(ValueError):
        _ = model.score(x, y)
    with pytest.raises(ValueError):
        _ = model.log_prob(x, y)
    with pytest.raises(ValueError):
        _ = model.sample_n(n=10, mu_given_x=x)
    with pytest.raises(ValueError):
        _ = model.predict(x)



import numpy as np
from scipy.special import beta, digamma

from regmmd.models.base_model import EstimationModel


class BetaBase(EstimationModel):
    """Beta distribution with density function
    p(x) ~ x^(a - 1)*(1 - x)^(b-1)

    """

    def __init__(self, alpha: float = None, beta: float = None, random_state=None):
        # k and theta respectively
        self.alpha = alpha
        self.beta = beta

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=random_state)

    def log_prob(self, x: float):
        if self.alpha is None or self.beta is None:
            raise ValueError(
                "Both parameters need to be defined"
                + "to be able to calculate the log_prob"
            )

        log_Z = -np.log(beta(self.alpha, self.beta))
        log_alpha = (self.alpha - 1) * np.log(x)
        log_beta = (self.beta - 1) * np.log(1 - x)
        return log_Z + log_alpha + log_beta

    def sample_n(self, n: int):
        if self.alpha is None or self.beta is None:
            raise ValueError("Both parameters need to be defined to be able to sample")

        return self.rng.beta(a=self.alpha, b=self.beta, size=(n,))

    def _init_params(self, X):
        mean = X.mean(axis=0)
        var = X.std(axis=0) ** 2

        C = (mean * (1 - mean) / var) - 1
        if self.alpha is None:
            self.alpha = C * mean
        if self.beta is None:
            self.beta = C * (1 - mean)

        return self._get_params()

    def _alpha_grad(self, x):
        log_exp = np.log(x)
        log_beta_func = -digamma(self.alpha) + digamma(self.alpha + self.beta)
        return log_exp + log_beta_func

    def _beta_grad(self, x):
        log_exp = np.log(1 - x)
        log_beta_func = -digamma(self.beta) + digamma(self.alpha + self.beta)
        return log_exp + log_beta_func


class BetaA(BetaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(alpha=par_v, beta=par_c, random_state=random_state)

    def score(self, x):
        if self.alpha is None or self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )

        return self._alpha_grad(x)

    def update(self, par_v):
        self.alpha = par_v

    def _get_params(self):
        par_v = self.alpha
        par_c = self.beta
        return par_v, par_c

    def _project_params(self, par_v):
        par_v = max(1e-6, par_v)
        return par_v


class BetaB(BetaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(alpha=par_c, beta=par_v, random_state=random_state)

    def score(self, x):
        if self.alpha is None or self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )
        return self._beta_grad(x)

    def update(self, par_v):
        self.beta = par_v

    def _get_params(self):
        par_c = self.alpha
        par_v = self.beta
        return par_v, par_c

    def _project_params(self, par_v):
        par_v = max(1e-6, par_v)
        return par_v


class Beta(BetaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        if par_v is None:
            super().__init__(alpha=None, beta=None, random_state=random_state)
        else:
            super().__init__(alpha=par_v[0], beta=par_v[1], random_state=random_state)

    def score(self, x):
        if self.alpha is None or self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )
        _alpha_grad = self._alpha_grad(x)
        _beta_grad = self._beta_grad(x)
        return np.array([_alpha_grad, _beta_grad]).T

    def update(self, par_v):
        self.alpha = par_v[0]
        self.beta = par_v[1]

    def _get_params(self):
        par_v = np.array([self.alpha, self.beta])
        par_c = None
        return par_v, par_c

    def _project_params(self, par_v):
        par_v[0] = max(1e-6, par_v[0])
        par_v[1] = max(1e-6, par_v[1])
        return par_v



from regmmd.models.regression.linear_gaussian import LinearGaussian, LinearGaussianLoc
from regmmd.models.regression.logistic import Logistic
from regmmd.models.regression.gamma import GammaRegression, GammaRegressionLoc
from regmmd.models.regression.poisson import PoissonRegression

__all__ = [
    "LinearGaussian",
    "LinearGaussianLoc",
    "Logistic",
    "GammaRegression",
    "GammaRegressionLoc",
    "PoissonRegression",
]



from regmmd.models.estimation import *
from regmmd.models.regression import *

from regmmd.models.estimation import __all__ as __all_estimation__
from regmmd.models.regression import __all__ as __all_regression__

__all__ = __all_estimation__ + __all_regression__



import numpy as np
from numpy.typing import NDArray
from scipy.special import gamma, digamma

from regmmd.models.base_model import RegressionModel
from sklearn.linear_model import GammaRegressor


class GammaRegressionBase(RegressionModel):
    """Gamma regression, where exp(X^T \\beta) = mean(Y). The
    mean parametrized gamma density is given by
    p(y | x) ~ (shape * y / mean)^shape exp(-shape * y / mean) /  y for a: shape, b: rate

    """

    def __init__(self, beta: float = None, shape: float = None, random_state=None):
        self.beta = beta
        self.shape = shape

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=random_state)

    def log_prob(self, X: NDArray, y: NDArray) -> NDArray:
        if self.shape is None or self.beta is None:
            raise ValueError(
                "Both parameters need to be defined"
                + "to be able to calculate the log_prob"
            )
        mu_given_x = self.predict(X)
        log_Z = -np.log(gamma(self.shape))
        log_y = (self.shape - 1) * np.log(y)
        log_shape = self.shape * np.log(self.shape)
        log_mu = -self.shape * np.log(mu_given_x)
        log_exp = -self.shape * y / mu_given_x
        return log_Z + log_y + log_shape + log_mu + log_exp

    def sample_n(self, n: int, mu_given_x: NDArray):
        if self.shape is None or self.beta is None:
            raise ValueError("Both parameters need to be defined to be able to sample")

        return self.rng.gamma(
            shape=self.shape, scale=mu_given_x / self.shape, size=(n,)
        )

    def predict(self, X):
        return np.exp(X @ self.beta)

    def _init_params(self, X, y):
        init_model = GammaRegressor(fit_intercept=False).fit(X, y)
        if self.beta is None:
            self.beta = init_model.coef_

        # Var(Y) = mu^2 / shape, -> shape =mu^2 / var(Y)
        if self.shape is None:
            y_hat = init_model.predict(X)
            mu_2 = np.mean(y_hat) ** 2
            var_y = np.var(y)
            self.shape = mu_2 / var_y

        return self._get_params()

    def _shape_grad(self, X: NDArray, y: NDArray) -> NDArray:
        mu_given_x = self.predict(X)

        log_gamma = -digamma(self.shape)
        log_y = np.log(y)
        log_shape = np.log(self.shape) + 1
        log_mu = -np.log(mu_given_x)
        log_exp = -y / mu_given_x
        return (log_gamma + log_y + log_shape + log_mu + log_exp)[:, np.newaxis]

    def _beta_grad(self, X: NDArray, y: NDArray) -> NDArray:
        mu_given_x = self.predict(X)

        residuals = self.shape * (y / mu_given_x - 1)
        return X * (residuals[:, np.newaxis])


class GammaRegressionLoc(GammaRegressionBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(beta=par_v, shape=par_c, random_state=random_state)

    def score(self, X, y):
        return self._beta_grad(X, y)

    def update(self, par_v):
        self.beta = par_v

    def _get_params(self):
        par_v = self.beta
        par_c = self.shape
        return par_v, par_c

    def _project_params(self, par_v):
        return par_v


class GammaRegression(GammaRegressionBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        if par_v is None:
            super().__init__(beta=None, shape=None, random_state=random_state)
        else:
            super().__init__(
                beta=par_v[:-1], shape=par_v[-1], random_state=random_state
            )

    def score(self, X, y):
        _beta_grad = self._beta_grad(X, y)
        _shape_grad = self._shape_grad(X, y)
        return np.hstack((_beta_grad, _shape_grad))

    def update(self, par_v):
        self.beta = par_v[:-1]
        self.shape = par_v[-1]

    def _get_params(self):
        par_v = np.concatenate((self.beta, np.array([self.shape])))
        par_c = None
        return par_v, par_c

    def _project_params(self, par_v):
        # par_v = max(0.5 , par_v) This was found empirically in the development of the R package
        par_v[-1] = max(1e-6, par_v[-1])
        return par_v



import numpy as np

from regmmd.models.base_model import EstimationModel


class GaussianBase(EstimationModel):
    def __init__(self, loc=None, scale=None, random_state=None):
        self.loc = loc
        self.scale = scale

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=self.random_state)

    def log_prob(self, x):
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Both parameters need to be defined"
                + "to be able to calculate the log_prob"
            )
        log_Z = -0.5 * np.log(2 * np.pi)
        log_sigma = -np.log(self.scale)
        log_exp = -((x - self.loc) ** 2 / (2 * self.scale**2))
        return log_Z + log_sigma + log_exp

    def sample_n(self, n):
        if self.loc is None or self.scale is None:
            raise ValueError("Both parameters need to be defined to be able to sample")

        return self.rng.normal(loc=self.loc, scale=self.scale, size=(n,))

    def _init_params(self, X):
        if self.loc is None:
            self.loc = np.median(X)
        if self.scale is None:
            self.scale = (5 / 4) * np.median(abs(X - np.median(X)))
        return self._get_params()

    def _loc_grad(self, x):
        return (x - self.loc) / (self.scale**2)

    def _scale_grad(self, x):
        return -1 / self.scale + (x - self.loc) ** 2 / (self.scale**3)


class GaussianLoc(GaussianBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(loc=par_v, scale=par_c, random_state=random_state)

    def score(self, x):
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )

        return self._loc_grad(x)

    def update(self, par_v):
        self.loc = par_v

    def _get_params(self):
        par_v = self.loc
        par_c = self.scale
        return par_v, par_c

    def _project_params(self, par_v):
        return par_v

    def _exact_fit(self, X, par_v, par_c, solver, kernel, bandwidth):
        if kernel == "Gaussian":
            from regmmd.optimizers import _gd_gaussian_loc_exact_estimation

            return _gd_gaussian_loc_exact_estimation(
                X=X,
                par_v=par_v,
                par_c=par_c,
                burn_in=solver["burnin"],
                n_step=solver["n_step"],
                stepsize=solver["stepsize"],
                bandwidth=bandwidth,
                epsilon=solver["epsilon"],
            )
        return None


class GaussianScale(GaussianBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(loc=par_c, scale=par_v, random_state=random_state)

    def score(self, x):
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )

        return self._scale_grad(x)

    def update(self, par_v):
        self.scale = par_v

    def _get_params(self):
        par_v = self.scale
        par_c = self.loc
        return par_v, par_c

    def _project_params(self, par_v):
        return max(1e-6, par_v)


class Gaussian(GaussianBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        if par_v is None:
            super().__init__(loc=None, scale=None, random_state=random_state)
        else:
            super().__init__(loc=par_v[0], scale=par_v[1], random_state=random_state)

    def score(self, x):
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Both parameters need to be defined to be able to calculate the score"
            )

        _score_loc = self._loc_grad(x)
        _score_scale = self._scale_grad(x)

        return np.array([_score_loc, _score_scale]).T

    def update(self, par_v):
        self.loc = par_v[0]
        self.scale = par_v[1]

    def _get_params(self):
        par_v = np.array([self.loc, self.scale])
        par_c = None
        return par_v, par_c

    def _project_params(self, par_v):
        par_v[1] = max(1e-6, par_v[1])
        return par_v



import numpy as np
import pytest

from regmmd.models import (
    Gaussian,
    GaussianLoc,
    GaussianScale,
    Beta,
    BetaA,
    BetaB,
    Gamma,
    GammaShape,
    GammaRate,
    Binomial,
    Poisson,
)
from regmmd.models import __all_estimation__
from regmmd.models.base_model import EstimationModel

RNG = np.random.default_rng(42)
X_GAUSS = RNG.normal(2.0, 1.0, size=(50,))
X_BETA = RNG.beta(2.0, 5.0, size=(50,))
X_GAMMA = RNG.gamma(shape=2.0, scale=1 / 1.5, size=(50,))
X_POISSON = RNG.poisson(lam=2, size=(50,))
X_BIN = RNG.binomial(n=10, p=0.7, size=(50,))


# --- BaseModel ---


def test_base_model_raises_error():
    class NewModel(EstimationModel):
        def __init__(self, par_v=None, par_c=None, random_state=None):
            self.par_v = par_v
            self.par_c = par_c
            self.random_state = random_state

    with pytest.raises(TypeError):
        model = NewModel()


# --- GaussianLoc ---


def test_gaussian_loc_init_params():
    model = GaussianLoc(random_state=0)
    model._init_params(X_GAUSS)
    assert model.loc is not None
    assert model.scale is not None


def test_gaussian_loc_sample_n():
    model = GaussianLoc(par_v=2.0, par_c=1.0, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)


def test_gaussian_loc_log_prob():
    model = GaussianLoc(par_v=2.0, par_c=1.0)
    result = model.log_prob(X_GAUSS)
    assert result.shape == X_GAUSS.shape


def test_gaussian_loc_score():
    model = GaussianLoc(par_v=2.0, par_c=1.0)
    score = model.score(X_GAUSS)
    assert score.shape == X_GAUSS.shape


def test_gaussian_loc_update():
    model = GaussianLoc(par_v=0.0, par_c=1.0)
    model.update(5.0)
    assert model.loc == 5.0


# --- GaussianScale ---


def test_gaussian_scale_project_params():
    model = GaussianScale(par_v=1.0, par_c=0.0)
    result = model._project_params(-1.0)
    assert result >= 1e-6


def test_gaussian_scale_score():
    model = GaussianScale(par_v=1.0, par_c=2.0)
    score = model.score(X_GAUSS)
    assert score.shape == X_GAUSS.shape


# --- Gaussian (both params) ---


def test_gaussian_sample_n():
    model = Gaussian(par_v=np.array([2.0, 1.0]))
    samples = model.sample_n(30)
    assert samples.shape == (30,)


def test_gaussian_score_shape():
    model = Gaussian(par_v=np.array([2.0, 1.0]))
    score = model.score(X_GAUSS)
    assert score.shape == (len(X_GAUSS), 2)


# --- Beta ---


def test_beta_log_prob():
    model = Beta(par_v=np.array([1.5, 2.5]))
    result = model.log_prob(X_BETA)
    assert result.shape == X_BETA.shape


def test_beta_both_score():
    model = Beta(par_v=np.array([2.0, 5.0]))
    score = model.score(X_BETA)
    assert score.shape == (len(X_BETA), 2)


def test_beta_inits():
    model = Beta(par_v=None, par_c=None)
    par_v, par_c = model._init_params(X_BETA)
    assert isinstance(par_v[0], float)
    assert isinstance(par_v[1], float)
    assert par_c is None


def test_beta_updates():
    model = Beta(par_v=np.array([2.0, 1.5]))
    model.update(par_v=np.array([3.0, 2.5]))
    assert model.alpha == 3.0
    assert model.beta == 2.5


def test_beta_projects():
    model = Beta(par_v=np.array([2.0, 1.5]))
    par_v = model._project_params(par_v=np.array([0.0, 0.0]))
    assert np.all(par_v > 0)


# --- BetaA ---


def test_betaA_init_params():
    model = BetaA(par_c=5.0, random_state=0)
    model._init_params(X_BETA)
    assert model.alpha is not None


def test_betaA_sample_n():
    model = BetaA(par_v=2.0, par_c=5.0, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)
    assert np.all((samples > 0) & (samples < 1))


def test_betaA_score():
    model = BetaA(par_v=2.0, par_c=5.0)
    score = model.score(X_BETA)
    assert score.shape == X_BETA.shape


def test_betaA_projects():
    model = BetaA(par_v=2.0, par_c=5.0)
    par_v = model._project_params(par_v=0)
    assert par_v > 0


# --- BetaB ---


def test_betaB_init_params():
    model = BetaA(par_c=5.0, random_state=0)
    model._init_params(X_BETA)
    assert model.alpha is not None


def test_betaB_score():
    model = BetaB(par_v=2.0, par_c=5.0)
    score = model.score(X_BETA)
    assert score.shape == X_BETA.shape


def test_betaB_projects():
    model = BetaB(par_v=2.0, par_c=5.0)
    par_v = model._project_params(par_v=0)
    assert par_v > 0


# --- Gamma---


def test_gamma_log_prob():
    model = Gamma(par_v=np.array([2.0, 1.0]), par_c=None)
    result = model.log_prob(X_GAMMA)
    assert result.shape == X_GAUSS.shape


def test_gamma_init():
    model = Gamma(par_v=None, par_c=None)
    model._init_params(X_GAMMA)
    assert model.rate is not None
    assert model.shape is not None


def test_gamma_both_score():
    model = Gamma(par_v=np.array([2.0, 1.5]))
    score = model.score(X_GAMMA)
    assert score.shape == (len(X_GAMMA), 2)


def test_gamma_updates():
    model = Gamma(par_v=np.array([2.0, 1.5]), par_c=None)
    model.update(par_v=np.array([3.0, 2.5]))
    assert model.shape == 3.0
    assert model.rate == 2.5


def test_gamma_projects():
    model = Gamma(par_v=np.array([2.0, 1.5]), par_c=None)
    par_v = model._project_params(par_v=np.array([0.0, 0.0]))
    assert np.all(par_v > 0)


# --- GammaShape ---


def test_gamma_shape_init_params():
    model = GammaShape(par_c=2.0, random_state=0)
    model._init_params(X_GAMMA)
    assert model.shape is not None


def test_gamma_shape_sample_n():
    model = GammaShape(par_v=1.5, par_c=2.0, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)
    assert np.all(samples > 0)


def test_gamma_shape_score():
    model = GammaShape(par_v=1.5, par_c=2.0)
    score = model.score(X_GAMMA)
    assert score.shape == X_GAMMA.shape


def test_gamma_shape_updates():
    model = GammaShape(par_v=1.5, par_c=2.0)
    model.update(par_v=2)
    assert model.shape == 2


def test_gamma_shape_projects():
    model = GammaShape(par_v=1.5, par_c=2.0)
    par_v = model._project_params(par_v=0)
    assert par_v > 0


# --- GammaRate ---


def test_gamma_rate_init_params():
    model = GammaRate(par_c=2.0, random_state=0)
    model._init_params(X_GAMMA)
    assert model.rate is not None


def test_gamma_rate_sample_n():
    model = GammaRate(par_v=1.5, par_c=2.0, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)
    assert np.all(samples > 0)


def test_gamma_rate_score():
    model = GammaRate(par_v=1.5, par_c=2.0)
    score = model.score(X_GAMMA)
    assert score.shape == X_GAMMA.shape


def test_gamma_rate_updates():
    model = GammaRate(par_v=1.5, par_c=2.0)
    model.update(par_v=2)
    assert model.rate == 2


def test_gamma_rate_projects():
    model = GammaRate(par_v=1.5, par_c=2.0)
    par_v = model._project_params(par_v=0)
    assert par_v > 0


# --- Binomial ---


def test_binomial_log_prob():
    model = Binomial(par_v=0.6, par_c=10)
    result = model.log_prob(X_BIN)
    assert result.shape == X_BIN.shape


def test_binomial_sample_n():
    model = Binomial(par_v=0.3, par_c=10, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)
    assert np.all((samples >= 0) & (samples <= 10))


def test_binomial_project_params():
    model = Binomial(par_v=0.5, par_c=10)
    assert model._project_params(0.0) >= 1e-6
    assert model._project_params(1.0) <= 1 - 1e-6


def test_binomial_score():
    model = Binomial(par_v=0.3, par_c=10)
    x = np.array([2, 5, 8], dtype=float)
    score = model.score(x)
    assert score.shape == x.shape


def test_binomial_init():
    model = Binomial(par_v=None, par_c=10)
    model._init_params(X_BIN)
    assert model.p is not None


def test_binomial_updates():
    model = Binomial(par_v=0.5, par_c=10)
    model.update(par_v=0.6)
    assert model.p == 0.6


# --- Poisson ---


def test_poisson_sample_n():
    model = Poisson(par_v=0.3, random_state=0)
    samples = model.sample_n(20)
    assert samples.shape == (20,)
    assert np.all((samples >= 0))


def test_poisson_project_params():
    model = Poisson(par_v=0.5)
    assert model._project_params(0.0) >= 1e-6


def test_poisson_init_params():
    model = Poisson()
    model._init_params(X_POISSON)
    assert model.lam is not None


def test_poisson_log_prob():
    model = Poisson(par_v=0.3)
    x = np.array([2, 5, 8], dtype=float)
    score = model.log_prob(x)
    assert score.shape == x.shape


def test_poisson_score():
    model = Poisson(par_v=0.3)
    x = np.array([2, 5, 8], dtype=float)
    score = model.score(x)
    assert score.shape == x.shape


def test_poisson_updates():
    model = Poisson(par_v=1.0)
    before = model.lam

    model.update(par_v=2)
    after = model.lam
    assert before == 1.0
    assert after == 2.0


# --- Parametrized test ---


@pytest.mark.parametrize("model", __all_estimation__)
def test_models_no_par_raises(model):
    model = eval(model)(par_v=None)
    x = np.array([2, 5, 8], dtype=float)
    with pytest.raises(ValueError):
        _ = model.score(x)
    with pytest.raises(ValueError):
        _ = model.log_prob(x)
    with pytest.raises(ValueError):
        _ = model.sample_n(x)



import numpy as np
import pytest

from regmmd.models.regression.linear_gaussian import LinearGaussian, LinearGaussianLoc
from regmmd.models.regression.gamma import GammaRegressionLoc
from regmmd.regression import MMDRegressor, _preprocess_data, NotFittedError

EXACT_SOLVER = {"n_step": 200, "stepsize": 0.5}

RNG = np.random.default_rng(42)


def test_log_prob():
    X = np.array([[1., 2.], [3., 4.]])
    y = np.array([5.0, 6.0])
    beta = np.array([0.5, 0.5])
    phi = 1.0

    par_v_init = np.concatenate((beta, np.array([phi])))
    model = LinearGaussian(par_v=par_v_init)

    log_prob = model.log_prob(X, y)
    assert isinstance(log_prob, float | np.floating)


def test_predict():
    X = np.array([[1., 2.], [3., 4.]])
    beta = np.array([0.5, 0.5])
    phi = 1.0

    par_v_init = np.concatenate((beta, np.array([phi])))
    model = LinearGaussian(par_v=par_v_init)

    predictions = model.predict(X)
    expected = X @ model.beta
    assert np.allclose(predictions, expected)


def test_sample_n():
    X = np.array([[1., 2.], [3., 4.]])
    beta = np.array([0.5, 0.5])
    phi = 1.0

    par_v_init = np.concatenate((beta, np.array([phi])))
    model = LinearGaussian(par_v=par_v_init, random_state=0)
    samples = model.sample_n(2, X @ model.beta)
    assert samples.shape == (2,)


def test_score():
    X = np.array([[1., 2.], [3., 4.]])
    y = np.array([5.0, 6.0])
    beta = np.array([0.5, 0.5])
    phi = 1.0

    par_v_init = np.concatenate((beta, np.array([phi])))
    model = LinearGaussian(par_v=par_v_init)
    score = model.score(X, y)
    assert score.shape == (2, 3)


# --- _preprocess_data ---


def test_preprocess_data_centers_X():
    X = RNG.normal(5.0, 1.0, size=(20, 2))
    y = RNG.normal(size=(20,))
    X_orig_mean = X.mean(axis=0).copy()
    X_out, y_out, X_offset, X_scale = _preprocess_data(X.copy(), y.copy())
    assert np.allclose(X_offset, X_orig_mean)


def test_preprocess_data_adds_intercept_column():
    X = RNG.normal(size=(20, 2))
    y = RNG.normal(size=(20,))
    X_out, y_out, X_offset, X_scale = _preprocess_data(
        X.copy(), y.copy(), fit_intercept=True
    )
    assert X_out.shape == (20, 3)


def test_preprocess_data_no_intercept():
    X = RNG.normal(size=(20, 2))
    y = RNG.normal(size=(20,))
    X_out, y_out, X_offset, X_scale = _preprocess_data(
        X.copy(), y.copy(), fit_intercept=False
    )
    assert X_out.shape == (20, 2)


# --- MMDRegressor ---

SOLVER = {"burnin": 50, "n_step": 100, "stepsize": 0.1}


def _make_regressor():
    # 2 input features; after fit_intercept=True preprocessing → 3 columns in X
    par_v = np.zeros(3)
    model = LinearGaussianLoc(par_v=par_v, par_c=1.0, random_state=0)
    return MMDRegressor(model=model, par_v=par_v, par_c=np.array([1.0]), solver=SOLVER)


def test_mmd_regressor_fit_sets_beta():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    reg = _make_regressor()
    reg.fit(X, y)
    assert hasattr(reg, "beta_")
    assert reg.beta_.shape == (1, 2)


def test_mmd_regressor_fit_sets_intercept():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    reg = _make_regressor()
    reg.fit(X, y)
    assert hasattr(reg, "intercept_")


def test_mmd_regressor_not_fitted_raises():
    par_v = np.zeros(3)
    model = LinearGaussianLoc(par_v=par_v, par_c=1.0)
    reg = MMDRegressor(model=model, par_v=par_v, solver=SOLVER)
    X_test = RNG.normal(size=(5, 2))
    with pytest.raises(NotFittedError):
        reg.predict(X_test)


@pytest.mark.parametrize("fit_intercept", [False, True])
def test_mmd_regressor_fitted_hat_predicts(fit_intercept):
    X = RNG.normal(size=(100, 2))
    noise = RNG.normal(size=(100,))
    beta = np.array([1, 2])
    if fit_intercept:
        y = X @ beta + 0.1 * noise + 1
    else:
        y = X @ beta + 0.1 * noise

    par_v = np.array([1.2, 2.1])
    model = LinearGaussianLoc(par_v=par_v, par_c=0.1)
    reg = MMDRegressor(
        model=model, par_v=par_v, solver=SOLVER, fit_intercept=fit_intercept
    )
    reg.fit(X, y)
    y_hat = reg.predict(X)
    mse = np.mean((y - y_hat) ** 2)
    assert mse < 0.5


@pytest.mark.parametrize("fit_intercept", [False, True])
def test_mmd_regressor_fitted_tilde_predicts(fit_intercept):
    # Small beta keeps mu = exp(X * beta) close to 1 and Var(Y|X) = mu^2/shape ~ 1.
    X = RNG.normal(size=(200, 2))
    beta = np.array([0.1, 0.2])
    shape = 20.0
    if fit_intercept:
        X_intercept = np.hstack((X, np.ones((X.shape[0], 1))))
        model = GammaRegressionLoc(
            par_v=np.concatenate((beta, np.array([1]))), par_c=shape, random_state=123
        )
        mu_given_x = model.predict(X=X_intercept)
        y = model.sample_n(n=200, mu_given_x=mu_given_x)
    else:
        model = GammaRegressionLoc(par_v=beta, par_c=shape, random_state=123)
        mu_given_x = model.predict(X=X)
        y = model.sample_n(n=200, mu_given_x=mu_given_x)

    par_v = np.array([0.15, 0.25])
    model = GammaRegressionLoc(par_v=par_v, par_c=shape, random_state=123)
    reg = MMDRegressor(
        model=model,
        par_v=par_v,
        par_c=shape,
        solver=SOLVER,
        fit_intercept=fit_intercept,
        bandwidth_X=0,
    )
    reg.fit(X, y)
    y_hat = reg.predict(X)
    mse = np.mean((y - y_hat) ** 2)
    assert mse < 0.5


# --- _exact_fit ---


def test_linear_gaussian_loc_exact_fit_tilde_returns_result():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.zeros(2)
    model = LinearGaussianLoc(par_v=par_v, par_c=1.0)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=1.0,
        solver=EXACT_SOLVER, kernel_y="Gaussian", bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X=0,
    )
    assert res is not None
    assert "estimator" in res


@pytest.mark.parametrize("kernel_y", ["Laplace", "Cauchy"])
def test_linear_gaussian_loc_exact_fit_non_gaussian_kernel_returns_none(kernel_y):
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.zeros(2)
    model = LinearGaussianLoc(par_v=par_v, par_c=1.0)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=1.0,
        solver=EXACT_SOLVER, kernel_y=kernel_y, bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X=0,
    )
    assert res is None


def test_linear_gaussian_loc_exact_fit_hat_returns_none():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.zeros(2)
    model = LinearGaussianLoc(par_v=par_v, par_c=1.0)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=1.0,
        solver=EXACT_SOLVER, kernel_y="Gaussian", bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X="auto",
    )
    assert res is None


def test_linear_gaussian_exact_fit_tilde_returns_result():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.array([0.0, 0.0, 1.0])  # [beta_0, beta_1, phi]
    model = LinearGaussian(par_v=par_v)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=None,
        solver=EXACT_SOLVER, kernel_y="Gaussian", bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X=0,
    )
    assert res is not None
    assert "estimator" in res
    assert res["estimator"].shape == (3,)


def test_linear_gaussian_exact_fit_non_gaussian_kernel_returns_none():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.array([0.0, 0.0, 1.0])
    model = LinearGaussian(par_v=par_v)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=None,
        solver=EXACT_SOLVER, kernel_y="Laplace", bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X=0,
    )
    assert res is None


def test_linear_gaussian_exact_fit_hat_returns_none():
    X = RNG.normal(size=(50, 2))
    y = RNG.normal(size=(50,))
    par_v = np.array([0.0, 0.0, 1.0])
    model = LinearGaussian(par_v=par_v)
    res = model._exact_fit(
        X=X, y=y, par_v=par_v, par_c=None,
        solver=EXACT_SOLVER, kernel_y="Gaussian", bandwidth_y=1.0,
        kernel_X="Laplace", bandwidth_X="auto",
    )
    assert res is None


def test_mmd_regressor_model_str_inits():
    reg = MMDRegressor(model="linear-gaussian", solver=SOLVER)


def test_mmd_regressor_wrong_model_str_raises():
    par_v = np.zeros(3)
    with pytest.raises(ValueError):
        reg = MMDRegressor(model="not-defined", par_v=par_v, solver=SOLVER)


def test_mmd_regressor_wrong_model_type_raises():
    par_v = np.zeros(3)
    with pytest.raises(TypeError):
        reg = MMDRegressor(model=None, par_v=par_v, solver=SOLVER)


def test_mmd_regressor_inits_params():
    reg = MMDRegressor(model="linear-gaussian-loc", solver=SOLVER, fit_intercept=True)
    par_v_before = reg.par_v
    par_c_before = reg.par_c

    X = np.array([[1], [3], [1.5], [2]], dtype=float)
    y = np.array([5.0, 6.0, 5.25, 5.5])

    reg.fit(X, y)
    par_v_after = reg.par_v
    par_c_after = reg.par_c

    assert par_v_before is None
    assert par_c_before is None
    assert par_v_after.shape == (2,)
    assert isinstance(par_c_after, float)



import numpy as np
from numpy.typing import NDArray
from scipy.spatial.distance import pdist

from regmmd.kernels import K1d_dist


def _median_heuristic(X: NDArray) -> float:
    """Compute the median heuristic for kernel bandwidth selection.

    Estimates the bandwidth as the median of pairwise Euclidean distances divided
    by sqrt(2), a common data-driven heuristic for kernel methods.

    Parameters
    ----------
    X : np.array, shape (n_samples,) or (n_samples, n_features)
        Input data.

    Returns
    -------
    median_dist : float
        Estimated bandwidth. Returns 1 if fewer than two samples are provided.
    """
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    pairwise_dists = pdist(X, metric="euclidean")
    if len(pairwise_dists) > 0:
        median_dist = np.median(pairwise_dists) / np.sqrt(2)
    else:
        median_dist = 1
    return median_dist


def sort_obs(X: NDArray) -> NDArray:
    """Sort all pairs of observations by their pairwise Euclidean distance.

    Computes pairwise distances between all rows of :math:`X` and returns the
    upper-triangular index pairs and corresponding distances, sorted from
    closest to most distant. Used to efficiently select the nearest pairs
    in the hat estimator gradient computation.

    Parameters
    ----------
    X : np.array, shape (n_samples, n_features)
        Input data whose rows are the observations to be paired.

    Returns
    -------
    result : dict
        Dictionary with two keys:

        - ``"DIST"`` : np.array of shape ``(n*(n-1)//2,)``, pairwise distances
          sorted in ascending order.
        - ``"IND"`` : np.array of shape ``(n*(n-1)//2, 2)``, row index pairs
          ``(i, j)`` with ``i < j`` corresponding to each distance in ``"DIST"``.
    """
    n = X.shape[0]
    dists = pdist(X, metric="euclidean")
    indices = np.triu_indices(n, k=1)
    indices = np.column_stack(indices)
    J = np.argsort(dists, axis=0)
    return {"DIST": dists[J], "IND": indices[J, :]}


def _get_grad_estimate(
    set_1: NDArray[np.int32],
    set_2: NDArray[np.int32],
    X: NDArray,
    K_X: NDArray,
    y_sampled_1: NDArray,
    y_sampled_2: NDArray,
    y: NDArray,
    model,
    kernel_y,
    bandwidth_y,
) -> NDArray:
    """Compute a partial gradient estimate for the hat estimator objective.

    Evaluates the gradient contribution from a specified subset of observation
    pairs :math:`(i, j)`. When ``set_1`` and ``set_2`` are provided, the gradient is
    weighted by the covariate kernel :math:`k_X` evaluated at those pairs. When both
    are ``None``, the diagonal term :math:`(i = j)` is computed without covariate
    kernel weighting.

    Parameters
    ----------
    set_1 : np.array[np.int32] or None
        Row indices of the first element of each pair. If ``None``, the
        diagonal :math:`(i = j)` contribution is computed.

    set_2 : np.array[np.int32] or None
        Row indices of the second element of each pair. If ``None``, the
        diagonal :math:`(i = j)` contribution is computed.

    X : np.array, shape (n_samples, n_features)
        Training input samples.

    K_X : np.array or None
        Precomputed covariate kernel values for the pairs defined by
        ``set_1`` and ``set_2``. Ignored when ``set_1`` is ``None``.

    y_sampled_1 : np.array, shape (n_samples,)
        First set of samples drawn from the model's conditional distribution.

    y_sampled_2 : np.array, shape (n_samples,)
        Second set of samples drawn from the model's conditional distribution.

    y : np.array, shape (n_samples,)
        Observed target values.

    model : RegressionModel
        The regression model. Must implement ``score``.

    kernel_y : str
        Kernel type applied to the target variable ``y``.

    bandwidth_y : float
        Bandwidth for the kernel applied to ``y``.

    Returns
    -------
    grad_estimate : np.array
        Gradient estimate contributions from the specified pairs. Shape is
        ``(n_params,)`` when ``set_1`` is ``None`` (diagonal term), or
        ``(len(set_1), n_params)`` for off-diagonal pairs.
    """
    if set_1 is not None and set_2 is not None:
        ker_sampled_1 = K1d_dist(
            y_sampled_1[set_1] - y_sampled_2[set_2],
            kernel=kernel_y,
            bandwidth=bandwidth_y,
        )
        ker_sampled_2 = K1d_dist(
            y_sampled_1[set_1] - y[set_2], kernel=kernel_y, bandwidth=bandwidth_y
        )
        ker = ker_sampled_1 - ker_sampled_2
        ker = K_X * ker

        grad_ll = model.score(X[set_1, :], y_sampled_1[set_1])
        grad_estimate = np.mean(ker @ grad_ll, axis=0)
    else:
        ker_sampled_1 = K1d_dist(
            y_sampled_1 - y_sampled_2, kernel=kernel_y, bandwidth=bandwidth_y
        )
        ker_sampled_2 = K1d_dist(
            y_sampled_1 - y, kernel=kernel_y, bandwidth=bandwidth_y
        )
        ker = ker_sampled_1 - ker_sampled_2

        grad_ll = model.score(X, y_sampled_1)
        grad_estimate = ker @ grad_ll

    return grad_estimate



from typing import Union

import numpy as np
from numpy.typing import NDArray

from regmmd.kernels import K1d, K1d_dist
from regmmd.models.base_model import EstimationModel, RegressionModel
from regmmd.utils import MMDResult

from regmmd.optimizers._common import _median_heuristic, sort_obs, _get_grad_estimate


def _sgd_estimation(
    X: NDArray,
    par_v: NDArray,
    par_c: NDArray,
    model: EstimationModel,
    kernel: str,
    burn_in: int = 500,
    n_step: int = 1000,
    stepsize: float = 1,
    bandwidth: float = 1.0,
    epsilon: float = 1e-4,
) -> MMDResult:
    """Estimate model parameters via AdaGrad stochastic gradient descent on the MMD criterion.

    Minimizes the MMD between the empirical distribution of ``X`` and the model's
    distribution using a stochastic gradient algorithm. The optimization runs a
    burn-in phase (without averaging) followed by a main phase with Polyak-Ruppert
    averaging of the iterates.

    Parameters
    ----------
    X : np.array, shape (n_samples,) or (n_samples, n_features)
        Observed data used to fit the model.

    par_v : np.array
        Initial values of the variable (optimized) model parameters.

    par_c : np.array
        Constant model parameters that are not optimized.

    model : EstimationModel
        The statistical model to fit. Must implement ``sample_n``, ``score``,
        ``update``, and ``_project_params``.

    kernel : str
        Kernel type used for the MMD computation. Supported options are
        ``"Gaussian"``, ``"Laplace"``, and ``"Cauchy"``.

    burn_in : int, default=500
        Number of burn-in iterations during which parameter iterates are not
        averaged.

    n_step : int, default=1000
        Number of averaging iterations following the burn-in phase.

    stepsize : float, default=1
        Initial step size for the AdaGrad update.

    bandwidth : float or str, default=1.0
        Bandwidth parameter for the kernel. If ``"auto"``, the bandwidth is
        selected using the median heuristic.

    epsilon : float, default=1e-4
        Initial accumulated squared gradient norm, used to stabilize the
        AdaGrad step size at the start of optimization.

    Returns
    -------
    res : MMDResult
        Dictionary containing:

        - ``par_v_init`` : initial variable parameters.
        - ``par_c_init`` : initial constant parameters.
        - ``stepsize`` : step size used.
        - ``bandwidth`` : bandwidth used (resolved if ``"auto"``).
        - ``estimator`` : Polyak-Ruppert average of parameter iterates.
        - ``trajectory`` : parameter trajectory of shape ``(n_params, burn_in + n_step + 1)``.
    """
    n = X.shape[0]

    if bandwidth == "auto":
        bandwidth = _median_heuristic(X)

    norm_grad = epsilon
    res = {
        "par_v_init": np.copy(par_v),
        "par_c_init": np.copy(par_c),
        "stepsize": stepsize,
        "bandwidth": bandwidth,
        "convergence": 1,
    }
    if len(par_v.shape) == 0:
        trajectory = np.zeros(shape=(1, burn_in + n_step + 1))
    else:
        trajectory = np.zeros(shape=(par_v.shape[0], burn_in + n_step + 1))
    trajectory[:, 0] = par_v

    for i in range(burn_in):
        x_sampled = model.sample_n(n=n)

        ker_sampled_1 = K1d(x_sampled, x_sampled, kernel=kernel, bandwidth=bandwidth)
        ker_sampled_1 = ker_sampled_1 / (n - 1)
        np.fill_diagonal(ker_sampled_1, 0)

        ker_sampled_2 = K1d(X, x_sampled, kernel=kernel, bandwidth=bandwidth) / n
        ker = ker_sampled_1 - ker_sampled_2

        grad_ll = model.score(x_sampled)
        grad = 2 * np.mean(ker @ grad_ll, axis=0)
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        norm_grad += np.sum(np.square(grad))
        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        model.update(par_v=par_v)

        # only for gaussian par[1] = max(par[1], 1 / (n ** 2))
        trajectory[:, i + 1] = par_v

    par_mean = par_v

    for i in range(n_step):
        x_sampled = model.sample_n(n=n)

        ker_sampled_1 = K1d(x_sampled, x_sampled, kernel=kernel, bandwidth=bandwidth)
        ker_sampled_1 = ker_sampled_1 / (n - 1)
        np.fill_diagonal(ker_sampled_1, 0)

        ker_sampled_2 = K1d(X, x_sampled, kernel=kernel, bandwidth=bandwidth) / n
        ker = ker_sampled_1 - ker_sampled_2

        grad_ll = model.score(x_sampled)
        grad = 2 * np.mean(ker @ grad_ll, axis=0)
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        norm_grad += np.sum(np.square(grad))
        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        par_mean = (par_mean * (i + 1) + par_v) / (i + 2)

        model.update(par_v=par_mean)
        trajectory[:, i + burn_in + 1] = par_mean

    res["estimator"] = par_mean
    res["trajectory"] = trajectory

    return res

def _sgd_hat_regression(
    X: NDArray,
    y: NDArray,
    par_v: NDArray,
    par_c: NDArray,
    model: RegressionModel,
    kernel_y: str,
    kernel_x: str = "Laplace",
    burn_in: int = 500,
    n_step: int = 1000,
    stepsize: float = 1,
    bandwidth_y: Union[float, str] = "auto",
    bandwidth_x: Union[float, str] = "auto",
    c_det: float = 0.2,
    c_rand: float = 0.1,
    epsilon: float = 1e-4,
    eps_sq: float = 1e-5,
    rng: np.random.Generator = np.random.default_rng(10),
) -> MMDResult:
    """Fit a regression model using the hat estimator via stochastic gradient descent,
    as described in `Universal Robust Regression via Maximum Mean Discrepancy`, Alquier,
    Gerber (2024).

    Minimizes the MMD objective using the product kernel :math:`k = k_X \\otimes k_Y`.
    The gradient is
    approximated efficiently by splitting pairs :math:`(X_i, X_j)` into three groups: all
    diagonal pairs, the `M_det` closest off-diagonal pairs (deterministic), and
    M_rand randomly selected distant pairs (stochastic). This yields a gradient
    estimator with cost linear in n per iteration.

    Compared to the tilde estimator, this estimator is robust to adversarial
    contamination of the training data, but is computationally more expensive due
    to the preprocessing of pairwise distances. See Alquier and Gerber (2024),
    Section 4.

    Parameters
    ----------
    X : np.arrau, shape (n_samples, n_features)
        Training input samples.

    y : np.array, shape (n_samples,)
        Target values.

    par_v : np.array
        Initial values of the variable (optimized) model parameters.

    par_c : np.array
        Constant model parameters that are not optimized.

    model : RegressionModel
        The regression model to fit. Must implement ``predict``, ``sample_n``,
        ``score``, ``update``, and ``_project_params``.

    kernel_y : str
        Kernel type applied to the target variable ``y``. Supported options are
        ``"Gaussian"``, ``"Laplace"``, and ``"Cauchy"``.

    kernel_x : str, default="Laplace"
        Kernel type applied to the covariates ``X``. Supported options are
        ``"Gaussian"``, ``"Laplace"``, and ``"Cauchy"``.

    burn_in : int, default=500
        Number of burn-in iterations during which parameter iterates are not
        included in the running average.

    n_step : int, default=1000
        Maximum number of averaging iterations following the burn-in phase.
        Early stopping is applied if the average gradient norm falls below
        ``eps_sq``.

    stepsize : float, default=1
        Initial step size for the AdaGrad update.

    bandwidth_y : float or str, default="auto"
        Bandwidth for the kernel applied to ``y``. If ``"auto"``, selected
        using the median heuristic.

    bandwidth_x : float or str, default="auto"
        Bandwidth for the kernel applied to ``X``. If ``"auto"``, selected
        using the median heuristic.

    c_det : float, default=0.2
        Fraction of n used to determine the number of deterministic
        (closest) off-diagonal pairs included per iteration.

    c_rand : float, default=0.1
        Fraction of n used to determine the number of randomly selected
        distant pairs included per iteration.

    epsilon : float, default=1e-4
        Initial accumulated squared gradient norm, used to stabilize the
        AdaGrad step size at the start of optimization.

    eps_sq : float, default=1e-5
        Convergence threshold for early stopping. Optimization stops when the
        log of the average gradient norm falls below ``log(eps_sq)``.

    rng : np.random.Generator, default=np.random.default_rng(10)
        Random number generator used for sampling distant pairs.

    Returns
    -------
    res : MMDResult
        Dictionary containing:

        - ``par_v_init`` : initial variable parameters.
        - ``par_c_init`` : initial constant parameters.
        - ``stepsize`` : step size used.
        - ``bandwidth_y`` : bandwidth used for ``y`` (resolved if ``"auto"``).
        - ``bandwidth_x`` : bandwidth used for ``X`` (resolved if ``"auto"``).
        - ``estimator`` : Polyak-Ruppert average of parameter iterates.
        - ``trajectory`` : cumulative-average parameter trajectory of shape
          ``(n_params, n_step_done)``.
    """
    n = X.shape[0]

    if bandwidth_x == "auto":
        bandwidth_x = _median_heuristic(X)

    if bandwidth_y == "auto":
        bandwidth_y = _median_heuristic(y)

    norm_grad = epsilon
    res = {
        "par_v_init": np.copy(par_v),
        "par_c_init": np.copy(par_c),
        "stepsize": stepsize,
        "bandwidth_y": bandwidth_y,
        "bandwidth_x": bandwidth_x,
        "convergence": 1,
    }
    trajectory = np.zeros(shape=(*par_v.shape, n_step + 1))
    trajectory[:, 0] = par_v
    grad_all = np.zeros(shape=par_v.shape)
    log_eps = np.log(eps_sq)

    # Precomputation stat
    sorted_obs = sort_obs(X)
    K_X = K1d_dist(sorted_obs["DIST"], kernel=kernel_x, bandwidth=bandwidth_x)
    M_det = int(np.floor(n * c_det))
    M_rand = max(int(np.floor(n * c_rand)), 1)
    l_KX = K_X.shape[0]
    if n + M_det + M_rand > l_KX:
        M_det = l_KX - n - 2
        M_rand = 2

    cons = ((n - 1) * n - 2 * M_det) / M_rand

    for i in range(burn_in):
        mu_given_x = model.predict(X)
        y_sampled_1 = model.sample_n(n, mu_given_x)
        y_sampled_2 = model.sample_n(n, mu_given_x)

        grad_p1 = _get_grad_estimate(
            set_1=None,
            set_2=None,
            X=X,
            K_X=None,
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        set_1 = sorted_obs["IND"][n : (n + M_det + 1), 0]
        set_2 = sorted_obs["IND"][n : (n + M_det + 1), 1]
        grad_p2 = _get_grad_estimate(
            set_1=set_1,
            set_2=set_2,
            X=X,
            K_X=K_X[n : n + M_det + 1],
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )

        use_X = rng.choice(np.arange(n + M_det, l_KX), size=M_rand, replace=False)
        set_1 = sorted_obs["IND"][use_X, 0]
        set_2 = sorted_obs["IND"][use_X, 1]
        grad_p3 = _get_grad_estimate(
            set_1=set_1,
            set_2=set_2,
            X=X,
            K_X=K_X[use_X],
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )

        grad = (
            grad_p1.sum(axis=0) + 2 * grad_p2.sum(axis=0) + cons * grad_p3.sum(axis=0)
        ) / n

        grad_all += grad
        norm_grad += np.sum(np.square(grad))

        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        model.update(par_v=par_v)

    for i in range(n_step):
        mu_given_x = model.predict(X)
        y_sampled_1 = model.sample_n(n, mu_given_x)
        y_sampled_2 = model.sample_n(n, mu_given_x)

        grad_p1 = _get_grad_estimate(
            set_1=None,
            set_2=None,
            X=X,
            K_X=None,
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        set_1 = sorted_obs["IND"][n : (n + M_det + 1), 0]
        set_2 = sorted_obs["IND"][n : (n + M_det + 1), 1]
        grad_p2 = _get_grad_estimate(
            set_1=set_1,
            set_2=set_2,
            X=X,
            K_X=K_X[n : n + M_det + 1],
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )

        use_X = rng.choice(np.arange(n + M_det, l_KX), size=M_rand, replace=False)
        set_1 = sorted_obs["IND"][use_X, 0]
        set_2 = sorted_obs["IND"][use_X, 1]
        grad_p3 = _get_grad_estimate(
            set_1=set_1,
            set_2=set_2,
            X=X,
            K_X=K_X[use_X],
            y_sampled_1=y_sampled_1,
            y_sampled_2=y_sampled_2,
            y=y,
            model=model,
            kernel_y=kernel_y,
            bandwidth_y=bandwidth_y,
        )

        grad = (
            grad_p1.sum(axis=0) + 2 * grad_p2.sum(axis=0) + cons * grad_p3.sum(axis=0)
        ) / n

        grad_all += grad
        norm_grad += np.sum(np.square(grad))

        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        model.update(par_v=par_v)
        # only for gaussian par[1] = max(par[1], 1 / (n ** 2))
        trajectory[:, i + 1] = par_v

        if np.isnan(np.mean(grad_all)):
            res["convergence"] = -1
            break

        g_1 = np.sqrt(np.sum(np.square(grad_all / (burn_in + i + 1)))) / par_v.shape
        if np.log(g_1) < log_eps:
            res["convergence"] = 0
            break

    n_step_done = int(i + 1)
    trajectory = trajectory[:, :n_step_done]
    trajectory = np.cumsum(trajectory, axis=1) / np.arange(1, n_step_done + 1)
    res["estimator"] = trajectory[:, -1]
    res["trajectory"] = trajectory

    return res


def _sgd_tilde_regression(
    X: NDArray,
    y: NDArray,
    par_v: NDArray,
    par_c: NDArray,
    model: RegressionModel,
    kernel: str,
    burn_in: int = 500,
    n_step: int = 1000,
    stepsize: float = 1,
    bandwidth: float = 1.0,
    epsilon: float = 1e-4,
    eps_sq: float = 1e-5,
) -> MMDResult:
    """Fit a regression model using the tilde estimator via stochastic gradient descent,
    as described in `Universal Robust Regression via Maximum Mean Discrepancy`, Alquier and
    Gerber (2024).

    Minimizes the MMD objective using only the kernel :math:`k_Y` on the target
    variable. The gradient is
    computed using pairs of samples drawn from the model's conditional distribution
    at the observed :math:`X` values, giving a cost linear in :math:`n` per iteration.

    Compared to the hat estimator, this estimator is computationally cheaper as it
    avoids the :math:`O(n^2)` preprocessing of pairwise covariate distances, but enjoys
    slightly weaker robustness guarantees, see Section 4, Alquier, Gerber (2024)

    Parameters
    ----------
    X : np.array, shape (n_samples, n_features)
        Training input samples.

    y : np.array, shape (n_samples,)
        Target values.

    par_v : np.array
        Initial values of the variable (optimized) model parameters.

    par_c : np.array
        Constant model parameters that are not optimized.

    model : RegressionModel
        The regression model to fit. Must implement ``predict``, ``sample_n``,
        ``score``, ``update``, and ``_project_params``.

    kernel : str
        Kernel type applied to the target variable ``y``. Supported options are
        ``"Gaussian"``, ``"Laplace"``, and ``"Cauchy"``.

    burn_in : int, default=500
        Number of burn-in iterations during which parameter iterates are not
        included in the running average.

    n_step : int, default=1000
        Maximum number of averaging iterations following the burn-in phase.
        Early stopping is applied if the average gradient norm falls below
        ``eps_sq``.

    stepsize : float, default=1
        Initial step size for the AdaGrad update.

    bandwidth : float or str, default=1.0
        Bandwidth for the kernel applied to ``y``. If ``"auto"``, selected
        using the median heuristic on ``X``.

    epsilon : float, default=1e-4
        Initial accumulated squared gradient norm, used to stabilize the
        AdaGrad step size at the start of optimization.

    eps_sq : float, default=1e-5
        Convergence threshold for early stopping. Optimization stops when the
        log of the average gradient norm falls below ``log(eps_sq)``.

    Returns
    -------
    res : MMDResult
        Dictionary containing:

        - ``par_v_init`` : initial variable parameters.
        - ``par_c_init`` : initial constant parameters.
        - ``stepsize`` : step size used.
        - ``bandwidth`` : bandwidth used (resolved if ``"auto"``).
        - ``estimator`` : Polyak-Ruppert average of parameter iterates.
        - ``trajectory`` : cumulative-average parameter trajectory of shape
          ``(n_params, n_step_done)``.
    """
    n = X.shape[0]

    if bandwidth == "auto":
        bandwidth = _median_heuristic(X)

    norm_grad = epsilon
    res = {
        "par_v_init": np.copy(par_v),
        "par_c_init": np.copy(par_c),
        "stepsize": stepsize,
        "bandwidth": bandwidth,
        "convergence": 1,
    }
    trajectory = np.zeros(shape=(*par_v.shape, n_step + 1))
    trajectory[:, 0] = par_v
    grad_all = np.zeros(shape=par_v.shape)
    log_eps = np.log(eps_sq)

    for i in range(burn_in):
        mu_given_x = model.predict(X)
        y_sampled_1 = model.sample_n(n, mu_given_x)
        y_sampled_2 = model.sample_n(n, mu_given_x)

        ker_sampled_1 = K1d_dist(
            y_sampled_1 - y_sampled_2, kernel=kernel, bandwidth=bandwidth
        )
        ker_sampled_2 = K1d_dist(y_sampled_1 - y, kernel=kernel, bandwidth=bandwidth)
        ker = ker_sampled_1 - ker_sampled_2

        grad_ll = model.score(X, y_sampled_1)
        grad = 2 * np.mean(ker @ grad_ll, axis=0)
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        grad_all += grad
        norm_grad += np.sum(np.square(grad))

        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        model.update(par_v=par_v)

    for i in range(n_step):
        mu_given_x = model.predict(X)
        y_sampled_1 = model.sample_n(n, mu_given_x)
        y_sampled_2 = model.sample_n(n, mu_given_x)

        ker_sampled_1 = K1d_dist(
            y_sampled_1 - y_sampled_2, kernel=kernel, bandwidth=bandwidth
        )
        ker_sampled_2 = K1d_dist(y_sampled_1 - y, kernel=kernel, bandwidth=bandwidth)
        ker = ker_sampled_1 - ker_sampled_2

        grad_ll = model.score(X, y_sampled_1)
        grad = 2 * np.mean(ker @ grad_ll, axis=0)
        # Expected outcome shape: (n, par.shape) -> (shape_par)

        grad_all += grad
        norm_grad += np.sum(np.square(grad))

        par_v -= stepsize * grad / np.sqrt(norm_grad)
        par_v = model._project_params(par_v=par_v)

        model.update(par_v=par_v)
        # only for gaussian par[1] = max(par[1], 1 / (n ** 2))
        trajectory[:, i + 1] = par_v

        if np.isnan(np.mean(grad_all)):
            res["convergence"] = -1
            break

        g_1 = np.sqrt(np.sum(np.square(grad_all / (burn_in + i + 1)))) / par_v.shape
        if np.log(g_1) < log_eps:
            res["convergence"] = 0
            break

    n_step_done = int(i + 1)
    trajectory = trajectory[:, :n_step_done]
    trajectory = np.cumsum(trajectory, axis=1) / np.arange(1, n_step_done + 1)
    res["estimator"] = trajectory[:, -1]
    res["trajectory"] = trajectory

    return res



from abc import ABC, abstractmethod
from typing import Dict, Union

from numpy.typing import NDArray
from regmmd.utils import MMDResult


class EstimationModel(ABC):
    @abstractmethod
    def sample_n(self, n: int) -> NDArray:
        """Generate n samples of the distribution with the initialized
        parameters of the distribution.

        Parameters
        ----------
        n : int, How many samples to generate
        """

    @abstractmethod
    def log_prob(self, x: NotADirectoryError) -> NDArray:
        """Evaluates to the log likelihood at the values x.

        Parameters
        ----------
        x : ndarray of shape (n_samples, n_features), the points to be evaluated
        """

    @abstractmethod
    def score(self, x) -> NDArray:
        """Evaluates to the gradient of the log likelihood with respect to the
        parameters at the values x.

        Parameters
        ----------
        x : ndarray of shape (n_samples, n_features), the points to be evaluated
        """

    @abstractmethod
    def _init_params(self, X) -> None:
        """Update the model with new parameters

        Parameters
        ----------
        """

    @abstractmethod
    def _get_params(self):
        """After named parameters are initialized, this function gets
        them in the order of par_v, par_c

        Parameters
        ----------
        """

    @abstractmethod
    def _project_params(self, par_v) -> NDArray:
        """Projection of potentially infeasible variable parameters to the
        feasible set

        Parameters
        ----------
        par_v : float, variable parameters
        """

    @abstractmethod
    def update(self, par_v) -> None:
        """Update the model with new parameters

        Parameters
        ----------
        par_v : float, variable parameters
        """

    def _exact_fit(
        self,
        X: NDArray,
        par_v: float,
        par_c: float,
        solver: Dict,
        kernel: str,
        bandwidth: Union[float, str],
    ) -> None | MMDResult:
        """Possible exact gradient descent optimization.

        Override in subclasses where a closed-form gradient exists for
        a specific kernel. Returns an MMDResult dict when an exact method
        is available, or None to fall back to SGD.
        """
        return None


class RegressionModel(EstimationModel):
    @abstractmethod
    def sample_n(self, n: int, mu_given_x: NDArray) -> NDArray:
        """Generate n samples of the distribution with the initialized
        parameters of the distribution and the conditional mean
        of the covariates

        Parameters
        ----------
        n : int, How many samples to generate

        mu_given_x: np.array, the covariates
        """

    @abstractmethod
    def predict(self, X: NDArray) -> NDArray:
        """Computes the mean of Y given X and the current parameters of the model

        Parameters
        ----------
        X : np.array
        """

    @abstractmethod
    def _init_params(self, X, y) -> None:
        """Update the model with new parameters

        Parameters
        ----------
        """

    def _exact_fit(
        self,
        X: NDArray,
        y: NDArray,
        par_v: NDArray,
        par_c: Union[float, None],
        solver: Dict,
        kernel_y: str,
        bandwidth_y: Union[float, str],
        kernel_X: str,
        bandwidth_X: Union[float, str],
    ) -> None | MMDResult:
        """Possible exact gradient descent optimization for regression.

        Override in subclasses where a closed-form gradient exists for
        a specific kernel. Returns an MMDResult dict when an exact method
        is available, or None to fall back to SGD.
        """
        return None



import numpy as np
from numpy.typing import NDArray

from regmmd.models.base_model import RegressionModel
from sklearn.linear_model import LinearRegression


class LinearGaussianBase(RegressionModel):
    def __init__(self, beta=None, phi=None, random_state=None):
        self.beta = beta
        self.phi = phi

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=self.random_state)

    def log_prob(self, X, y):
        if self.beta is None or self.phi is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the log_prob"
            )

        n = X.shape[0]

        log_Z = -n * 0.5 * np.log(2 * np.pi)
        log_sigma = -n * np.log(self.phi)
        log_exp = -np.sum(np.square(y - X @ self.beta)) / (2 * (self.phi))
        return log_Z + log_sigma + log_exp

    def sample_n(self, n: int, mu_given_x: NDArray) -> NDArray:
        if self.beta is None or self.phi is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the log_prob"
            )

        noise_sampled = self.rng.normal(loc=0, scale=np.sqrt(self.phi), size=(n,))
        return mu_given_x + noise_sampled

    def predict(self, X):
        """Outputs the mean given X, parameters need to be initialized for this"""
        if self.beta is None or self.phi is None:
            raise ValueError("Both parameters need to be defined to predict")

        return X @ self.beta

    def _beta_grad(self, X, y):
        mu = self.predict(X)

        residuals = (y - mu)[:, np.newaxis]
        score_beta = X * residuals / self.phi
        return score_beta

    def _phi_grad(self, X, y):
        mu = self.predict(X)

        residuals = (y - mu)[:, np.newaxis]
        score_phi = -1 / (2 * self.phi) + residuals**2 / (2 * (self.phi**2))
        return score_phi

    def _init_params(self, X, y):
        init_model = LinearRegression(fit_intercept=False).fit(X, y)
        if self.beta is None:
            self.beta = init_model.coef_

        if self.phi is None:
            y_hat = init_model.predict(X)
            phi_estimate = max(np.var(y_hat - y), 1e-6)
            self.phi = phi_estimate
        return self._get_params()


class LinearGaussian(LinearGaussianBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        if par_v is None:
            super().__init__(beta=None, phi=None, random_state=random_state)
        else:
            super().__init__(beta=par_v[:-1], phi=par_v[-1], random_state=random_state)

    def score(self, X, y):
        """gradient of the log-likelihood for each individual data point"""
        if self.beta is None or self.phi is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the score"
            )

        score_beta = self._beta_grad(X, y)
        score_phi = self._phi_grad(X, y)

        return np.hstack((score_beta, score_phi))

    def update(self, par_v):
        self.beta = par_v[:-1]
        self.phi = par_v[-1]

    def _project_params(self, par_v):
        par_v[-1] = max(1e-6, par_v[-1])
        return par_v

    def _get_params(self):
        par_v = np.concatenate((self.beta, np.array([self.phi])))
        par_c = None
        return par_v, par_c

    def _exact_fit(
        self,
        X,
        y,
        par_v,
        par_c,
        solver,
        kernel_y,
        bandwidth_y,
        kernel_X=None,
        bandwidth_X=None,
    ):
        if kernel_y == "Gaussian" and (bandwidth_X is None or bandwidth_X == 0):
            from regmmd.optimizers import _gd_backtracking_lg_tilde_regression

            return _gd_backtracking_lg_tilde_regression(
                X=X,
                y=y,
                par_v=par_v,
                n_step=solver["n_step"],
                stepsize=solver["stepsize"],
                bandwidth=bandwidth_y,
                alpha=solver.get("alpha", 0.8),
                eps_gd=solver.get("eps_gd", 1e-5),
            )
        return None


class LinearGaussianLoc(LinearGaussianBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(beta=par_v, phi=par_c, random_state=random_state)

    def score(self, X, y):
        """gradient of the log-likelihood for each individual data point"""
        if self.beta is None or self.phi is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the score"
            )

        score_beta = self._beta_grad(X, y)
        return score_beta

    def update(self, par_v):
        self.beta = par_v

    def _project_params(self, par_v):
        return par_v

    def _get_params(self):
        par_v = self.beta
        par_c = self.phi
        return par_v, par_c

    def _exact_fit(
        self,
        X,
        y,
        par_v,
        par_c,
        solver,
        kernel_y,
        bandwidth_y,
        kernel_X=None,
        bandwidth_X=None,
    ):
        if kernel_y == "Gaussian" and (bandwidth_X is None or bandwidth_X == 0):
            from regmmd.optimizers import _gd_backtracking_lg_loc_tilde_regression

            return _gd_backtracking_lg_loc_tilde_regression(
                X=X,
                y=y,
                par_v=par_v,
                par_c=par_c,
                n_step=solver["n_step"],
                stepsize=solver["stepsize"],
                bandwidth=bandwidth_y,
                alpha=solver.get("alpha", 0.8),
                eps_gd=solver.get("eps_gd", 1e-5),
            )
        return None



import numpy as np
from scipy.special import gamma, digamma

from regmmd.models.base_model import EstimationModel


class GammaBase(EstimationModel):
    """Gamma distribution with density function
    p(x) ~ x^(a - 1)exp(-bx) for a: shape, b: rate

    """

    def __init__(self, shape: float = None, rate: float = None, random_state=None):
        # k and theta respectively
        self.shape = shape
        self.rate = rate

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=random_state)

    def log_prob(self, x: float):
        if self.shape is None or self.rate is None:
            raise ValueError(
                "Both parameters need to be defined"
                + "to be able to calculate the log_prob"
            )

        log_Z = -np.log(gamma(x))
        log_exp = -x * self.rate
        log_x = (self.shape - 1) * np.log(x)
        log_rate = self.shape * np.log(self.rate)
        return log_Z + log_exp + log_x + log_rate

    def sample_n(self, n: int):
        if self.shape is None or self.rate is None:
            raise ValueError("Both parameters need to be defined to be able to sample")

        return self.rng.gamma(shape=self.shape, scale=1 / self.rate, size=(n,))

    def _init_params(self, X):
        mean = X.mean(axis=0)
        var = X.std(axis=0) ** 2

        if self.shape is None:
            self.shape = (mean**2) / var
        if self.rate is None:
            self.rate = mean / var

        return self._get_params()

    def _shape_grad(self, x):
        log_exp = np.log(x)
        log_rate = np.log(self.rate)
        log_gamma = digamma(self.shape)
        return log_exp + log_rate + log_gamma

    def _rate_grad(self, x):
        log_exp = -x
        log_rate = self.shape / self.rate
        return log_exp + log_rate


class GammaShape(GammaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(shape=par_v, rate=par_c, random_state=random_state)

    def score(self, x):
        if self.shape is None or self.rate is None:
            raise ValueError(
                "Both parameters need to be defined to be calculate the score"
            )
        return self._shape_grad(x)

    def update(self, par_v):
        self.shape = par_v

    def _get_params(self):
        par_v = self.shape
        par_c = self.rate
        return par_v, par_c

    def _project_params(self, par_v):
        par_v = max(1e-6, par_v)
        return par_v


class GammaRate(GammaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(shape=par_c, rate=par_v, random_state=random_state)

    def score(self, x):
        if self.shape is None or self.rate is None:
            raise ValueError(
                "Both parameters need to be defined to be calculate the score"
            )
        return self._rate_grad(x)

    def update(self, par_v):
        self.rate = par_v

    def _get_params(self):
        par_c = self.shape
        par_v = self.rate
        return par_v, par_c

    def _project_params(self, par_v):
        # par_v = max(0.5 , par_v) This was found empirically in the development of the R package
        par_v = max(1e-6, par_v)
        return par_v


class Gamma(GammaBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        if par_v is None:
            super().__init__(shape=None, rate=None, random_state=random_state)
        else:
            super().__init__(shape=par_v[0], rate=par_v[1], random_state=random_state)

    def score(self, x):
        if self.shape is None or self.rate is None:
            raise ValueError(
                "Both parameters need to be defined to be calculate the score"
            )
        _shape_grad = self._shape_grad(x)
        _rate_grad = self._rate_grad(x)
        return np.array([_shape_grad, _rate_grad]).T

    def update(self, par_v):
        self.shape = par_v[0]
        self.rate = par_v[1]

    def _get_params(self):
        par_c = self.shape
        par_v = self.rate
        return par_v, par_c

    def _project_params(self, par_v):
        # par_v = max(0.5 , par_v) This was found empirically in the development of the R package
        par_v[0] = max(1e-6, par_v[0])
        par_v[1] = max(1e-6, par_v[1])
        return par_v



from regmmd.optimizers._common import (
    _median_heuristic,
    sort_obs,
    _get_grad_estimate,
)
from regmmd.optimizers._sgd import (
    _sgd_estimation,
    _sgd_hat_regression,
    _sgd_tilde_regression,
)
from regmmd.optimizers._exact_estimation import (
    _gd_gaussian_loc_exact_estimation,
)
from regmmd.optimizers._exact_regression import (
    _gd_backtracking_lg_loc_tilde_regression,
    _gd_backtracking_lg_tilde_regression,
    _gd_backtracking_logistic_tilde_regression,
    _gd_exact_logistic_hat_regression,
)

__all__ = [
    "_median_heuristic",
    "sort_obs",
    "_get_grad_estimate",
    "_sgd_estimation",
    "_sgd_hat_regression",
    "_sgd_tilde_regression",
    "_gd_gaussian_loc_exact_estimation",
    "_gd_backtracking_lg_loc_tilde_regression",
    "_gd_backtracking_lg_tilde_regression",
    "_gd_backtracking_logistic_tilde_regression",
    "_gd_exact_logistic_hat_regression",
]



import numpy as np
from numpy.typing import NDArray

from regmmd.models.base_model import RegressionModel
from sklearn.linear_model import LogisticRegression


class LogisticBase(RegressionModel):
    def __init__(self, beta=None, random_state=None):
        self.beta = beta

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=self.random_state)

    def log_prob(self, X, y):
        if self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the log_prob"
            )

        mu = X @ self.beta
        p = self._link_func(mu)
        log_p_1 = np.log(p)
        log_p_2 = np.log(1 - p)
        return np.sum(y * log_p_1 + (1 - y) * log_p_2)

    def sample_n(self, n: int, mu_given_x: NDArray) -> NDArray:
        if self.beta is None:
            raise ValueError("Both parameters need to be defined to sample")

        y_sampled = self.rng.binomial(1, mu_given_x, size=(n,))
        return y_sampled

    def predict(self, X):
        if self.beta is None:
            raise ValueError("Both parameters need to be defined to predict")

        """Outputs the mean given X, parameters need to be initialized for this"""
        return self._link_func(X @ self.beta)

    def _link_func(self, mu):
        return 1 / (1 + np.exp(-mu))

    def _project_params(self, par_v):
        return par_v

    def _init_params(self, X, y):
        if self.beta is None:
            init_model = LogisticRegression(fit_intercept=False).fit(X, y)
            self.beta = init_model.coef_[0, :]
        return self._get_params()


class Logistic(LogisticBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(beta=par_v, random_state=random_state)

    def score(self, X, y):
        """gradient of the log-likelihood for each individual data point"""
        if self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the score"
            )
        p = self.predict(X)

        residuals = (y - p)[:, np.newaxis]
        score_beta = X * residuals

        return score_beta

    def update(self, par_v):
        self.beta = par_v

    def _get_params(self):
        par_v = self.beta
        par_c = None
        return par_v, par_c

    def _exact_fit(
        self,
        X,
        y,
        par_v,
        par_c,
        solver,
        kernel_y,
        bandwidth_y,
        kernel_X=None,
        bandwidth_X=None,
    ):
        if bandwidth_X is None or bandwidth_X == 0:
            from regmmd.optimizers import _gd_backtracking_logistic_tilde_regression

            return _gd_backtracking_logistic_tilde_regression(
                X=X,
                y=y,
                par_v=par_v,
                n_step=solver["n_step"],
                stepsize=solver["stepsize"],
                bandwidth=bandwidth_y,
                kernel=kernel_y,
                alpha=solver.get("alpha", 0.8),
                eps_gd=solver.get("eps_gd", 1e-5),
            )
        else:
            from regmmd.optimizers import _gd_exact_logistic_hat_regression

            return _gd_exact_logistic_hat_regression(
                X=X,
                y=y,
                par_v=par_v,
                kernel=kernel_y,
                kernel_x=kernel_X,
                burn_in=solver.get("burnin", 500),
                n_step=solver["n_step"],
                stepsize=solver["stepsize"],
                bandwidth_y=bandwidth_y,
                bandwidth_x=bandwidth_X,
                c_det=solver.get("c_det", 0.2),
                c_rand=solver.get("c_rand", 0.1),
                epsilon=solver.get("epsilon", 1e-4),
                eps_sq=solver.get("eps_sq", 1e-5),
            )



from typing import Dict, Optional, Union
from enum import Enum

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin

from regmmd.models import (
    LinearGaussianLoc,
    LinearGaussian,
    Logistic,
    GammaRegression,
    GammaRegressionLoc,
    PoissonRegression,
)
from regmmd.models.base_model import RegressionModel
from regmmd.optimizers import _sgd_hat_regression, _sgd_tilde_regression
from regmmd.utils import MMDResult

from sklearn.utils.validation import check_X_y, check_array


class DefinedModels(Enum):
    LINEAR_GAUSSIAN = LinearGaussian
    LOGISTIC = Logistic
    LINEAR_GAUSSIAN_LOC = LinearGaussianLoc
    GAMMA_REGRESSION = GammaRegression
    GAMMA_REGRESSION_LOC = GammaRegressionLoc
    POISSON_REGRESSION = PoissonRegression


def _preprocess_data(
    X,
    y,
    fit_intercept=True,
):
    """Common data preprocessing for fitting linear model, adapted
    to a condensed version from sklearn.linear_models.base

    If `fit_intercept=True` this preprocessing centers `X` and
    store the mean vector in `X_offset`.

    If `fit_intercept=False`, no centering is performed and `X_offset` is
    set to zero.

    Returns
    -------
    X_out : ndarray of shape (n_samples, n_features)
        If copy=True a copy of the input X is triggered, otherwise operations are
        inplace.
        If input X is dense, then X_out is centered.

    y_out : ndarray of shape (n_samples,) or (n_samples, n_targets)
        Centered version of y. Possibly performed inplace on input y depending
        on the copy_y parameter.

    X_offset : ndarray of shape (n_features,)
        The mean per column of input X.

    X_scale: float or ndarray of shape (n_features,)

    """
    n_samples, _ = X.shape

    X_offset = X.mean(axis=0)
    X -= X_offset

    X_scale = np.std(X, axis=0)[np.newaxis, :]
    X /= X_scale

    if fit_intercept:
        X = np.hstack((X, np.ones(shape=(n_samples, 1))))

    return X, y, X_offset, X_scale


class MMDRegressor(RegressorMixin, BaseEstimator):
    """Regression using the Maximum Mean Discrepancy (MMD) criterion.

    This class implements regression using the MMD criterion, which is a kernel-based
    method to compare distributions by measuring the distance between mean embeddings
    in a Reproducing Kernel Hilbert Space (RKHS).

    MMDRegressor fits a regression model by minimizing the MMD between the distributions
    of the observed data and the model's predictions. It supports various kernel types
    and bandwidth selection methods for both the input features and the target variables.

    Parameters
    ----------
    model : RegressionModel
        The statistical model used for regression, provided as an instance of a
        `RegressionModel` class with initialized parameters. This model defines the
        relationship between the input features and the target variable.

    fit_intercept : bool, default=True
        Specifies whether to calculate the intercept for the model. If set to `False`,
        the model assumes that the data is already centered, and no intercept will be
        fitted.

    par_v : np.array, optional
        Initial values for the variable parameters of the model. If `None`, the model
        will use default initial values.

    par_c : np.array, optional
        Initial values for the constant parameters of the model. If `None`, the model
        will use default initial values.

    kernel_y : str, default="Gaussian"
        The kernel type used for the target variable `y`. Supported options are
        "Gaussian", "Laplace", and "Cauchy".

    kernel_X : str, default="Laplace"
        The kernel type used for the input features `X`. Supported options are
        "Gaussian", "Laplace", and "Cauchy".

    bandwidth_y : Union[str, float], default="auto"
        The bandwidth parameter for the kernel applied to the target variable `y`.
        If set to "auto", the bandwidth is determined using a heuristic method,
        such as the median heuristic.

    bandwidth_X : Union[str, float], default="auto"
        The bandwidth parameter for the kernel applied to the input features `X`.
        If set to "auto", the bandwidth is determined using a heuristic method,
        such as the median heuristic.

    solver : dict, optional
        A dictionary specifying the solver parameters for the optimization process.
        It should include keys such as "burnin" (number of burn-in iterations), "n_step" (number of
        optimization steps), and "stepsize" (learning rate for the optimizer).
        If `None`, default solver settings are used.

    random_state : int, optional
        random seed to be passed to the model and any sampler used in the SGD optimizers.

    Attributes
    ----------
    X_offset : np.array or None
        The offset applied to the input features `X` during preprocessing. This is
        used when `fit_intercept` is `True`.

    y_offset : np.array or None
        The offset applied to the target variable `y` during preprocessing.

    X_scale : np.array or None
        The scale factor applied to the input features `X` during preprocessing.

    par_v : np.array
        The estimated variable parameters of the model after fitting.

    Notes
    -----
    - The `fit` method preprocesses the data, fits the model using the specified solver,
      and updates the model parameters.
    - The `predict` method uses the fitted model to make predictions on new data.
    """

    def __init__(
        self,
        model: DefinedModels | RegressionModel,
        fit_intercept: bool = True,
        par_v: Optional[np.array] = None,
        par_c: Optional[np.array] = None,
        kernel_y: str = "Gaussian",
        kernel_X: str = "Laplace",
        bandwidth_y: Union[str, float] = "auto",
        bandwidth_X: Union[str, float] = "auto",
        solver: Optional[Dict] = None,
        random_state: Optional[int] = None,
    ):
        self.fit_intercept = fit_intercept
        if isinstance(model, str):
            try:
                self.model = DefinedModels[model.upper().replace("-", "_")].value(
                    par_v=par_v, par_c=par_c, random_state=random_state
                )
            except KeyError:
                raise ValueError("model string is not defined by the package.")
        elif isinstance(model, RegressionModel):
            self.model = model
        else:
            raise TypeError("Expected either string or RegressionModel!")

        self.par_v = par_v
        self.par_c = par_c
        self.kernel_y = kernel_y
        self.kernel_X = kernel_X
        self.bandwidth_y = bandwidth_y
        self.bandwidth_X = bandwidth_X
        self.solver = solver

        self.X_offset = None
        self.y_offset = None
        self.X_scale = None

    def fit(self, X: np.ndarray, y: np.ndarray) -> MMDResult:
        """Fit the MMD regression model according to the given training data.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Training input samples.

        y : np.ndarray, shape (n_samples,)
            Target values.

        Returns
        -------
        res : MMDResult
            A dictionary containing the results of the optimization process, including
            the estimated parameters and the optimization trajectory.
        """
        X, y = self._validate_data(X, y)
        n_features = X.shape[1]

        y_int = y.astype(int)
        if (y_int - y).sum() != 0:
            X, y, X_offset, X_scale = _preprocess_data(
                X,
                y,
                fit_intercept=self.fit_intercept,
            )
            self.X_offset = X_offset
            self.X_scale = X_scale

        if (
            self.fit_intercept
            and self.model.beta is not None
            and len(self.model.beta) == n_features
        ):
            self.par_v = np.insert(self.par_v, n_features, 1)
            self.model.update(self.par_v)
            print("model.beta", self.model.beta)

        if self.par_v is None or self.par_c is None:
            self.par_v, self.par_c = self.model._init_params(X=X, y=y)

        res = self.model._exact_fit(
            X=X,
            y=y,
            par_v=self.par_v,
            par_c=self.par_c,
            solver=self.solver,
            kernel_y=self.kernel_y,
            bandwidth_y=self.bandwidth_y,
            kernel_X=self.kernel_X,
            bandwidth_X=self.bandwidth_X,
        )
        if res is None:
            if self.bandwidth_X == 0:
                res = _sgd_tilde_regression(
                    X=X,
                    y=y,
                    par_v=self.par_v,
                    par_c=self.par_c,
                    model=self.model,
                    kernel=self.kernel_y,
                    burn_in=self.solver["burnin"],
                    n_step=self.solver["n_step"],
                    stepsize=self.solver["stepsize"],
                    bandwidth=self.bandwidth_y,
                )
            else:
                res = _sgd_hat_regression(
                    X=X,
                    y=y,
                    par_v=self.par_v,
                    par_c=self.par_c,
                    model=self.model,
                    kernel_y=self.kernel_y,
                    kernel_x=self.kernel_X,
                    burn_in=self.solver["burnin"],
                    n_step=self.solver["n_step"],
                    stepsize=self.solver["stepsize"],
                    bandwidth_y=self.bandwidth_y,
                    bandwidth_x=self.bandwidth_X,
                )

        print("after model", self.model.beta)

        if not isinstance(self.model, Logistic):
            self.beta_ = res["estimator"][:n_features] / self.X_scale
            res["estimator"][:n_features] = self.beta_

            if self.fit_intercept:
                if self.beta_.ndim == 1:
                    self.intercept_ = (
                        res["estimator"][n_features] - self.X_offset @ self.beta_
                    )
                else:
                    self.intercept_ = (
                        res["estimator"][n_features] - self.X_offset @ self.beta_.T
                    )
                res["estimator"][n_features] = self.intercept_[0]
            else:
                self.intercept_ = 0.0
            self.par_v = res["estimator"]
        else:
            self.beta_ = res["estimator"][:n_features]
            if self.fit_intercept:
                self.intercept_ = res["estimator"][n_features]
            else:
                self.intercept_ = 0.0

        self.model.update(par_v=self.par_v)
        return res

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict using the MMD regression model.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Input samples for which to compute the predictions.

        Returns
        -------
        y_pred : np.ndarray, shape (n_samples,)
            The predicted target values.
        """
        X = self._validate_data(X)
        self._check_is_fitted()

        if self.fit_intercept:
            X = np.hstack((X, np.ones(shape=(X.shape[0], 1))))

        return self.model.predict(X)

    def _validate_data(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Union[np.ndarray, tuple]:
        """Validate input arrays X and y.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Input feature array.

        y : np.ndarray, shape (n_samples,), optional
            Target array.

        Returns
        -------
        X_validated : np.ndarray
            Validated input feature array.

        y_validated : np.ndarray, optional
            Validated target array, if provided.
        """
        if y is not None:
            X, y = check_X_y(X, y, multi_output=False, y_numeric=True)
            return X, y
        else:
            X = check_array(X)
            return X

    def _check_is_fitted(self) -> None:
        """Check if the model is fitted.

        Raises
        ------
        NotFittedError
            If the model is not fitted yet.
        """
        if not hasattr(self, "beta_"):
            raise NotFittedError(
                "This MMDRegressor instance is not fitted yet. Call 'fit' with appropriate arguments before using this method."
            )


class NotFittedError(ValueError):
    """Exception class to raise if the model is used before fitting."""

    pass



import numpy as np
import pytest

from regmmd.optimizers import (
    _median_heuristic,
    _sgd_estimation,
    _gd_gaussian_loc_exact_estimation,
    sort_obs,
)
from regmmd.utils import MMDResult
from regmmd.models import GaussianLoc, Gaussian

RNG = np.random.default_rng(42)


# --- _median_heuristic ---


def test_median_heuristic_1d():
    X = np.array([0.0, 1.0, 2.0, 3.0])
    result = _median_heuristic(X)
    assert isinstance(result, float | np.floating)
    assert result >= 0


def test_median_heuristic_2d():
    X = RNG.normal(size=(20, 2))
    result = _median_heuristic(X)
    assert isinstance(result, float | np.floating)
    assert result >= 0


def test_median_heuristic_single_point_is_one():
    X = np.array([[1.0, 2.0]])
    result = _median_heuristic(X)
    assert result == 1


# --- sort_obs ---


def test_sort_obs_keys():
    X = RNG.normal(size=(10, 2))
    result = sort_obs(X)
    assert "DIST" in result
    assert "IND" in result


def test_sort_obs_dist_is_sorted():
    X = RNG.normal(size=(15, 2))
    result = sort_obs(X)
    dists = result["DIST"]
    assert np.all(dists[:-1] <= dists[1:])


def test_sort_obs_ind_shape():
    n = 10
    X = RNG.normal(size=(n, 2))
    result = sort_obs(X)
    expected_pairs = n * (n - 1) // 2
    assert result["IND"].shape == (expected_pairs, 2)


# --- _gd_gaussian_loc_exact_estimation ---


def test_gd_gaussian_loc_exact_result_keys():
    X = RNG.normal(2.0, 1.0, size=(50,))
    res = _gd_gaussian_loc_exact_estimation(
        X, par_v=0.0, par_c=1.0, burn_in=10, n_step=20, stepsize=1.0, bandwidth=1.0
    )
    for key in (
        "par_v_init",
        "par_c_init",
        "stepsize",
        "estimator",
        "trajectory",
        "bandwidth",
    ):
        assert key in res


def test_gd_gaussian_loc_exact_trajectory_length():
    X = RNG.normal(2.0, 1.0, size=(50,))
    burn_in, n_step = 10, 20
    res = _gd_gaussian_loc_exact_estimation(
        X, par_v=0.0, par_c=1.0, burn_in=burn_in, n_step=n_step
    )
    assert res["trajectory"].shape == (burn_in + n_step + 1,)


def test_gd_gaussian_loc_exact_auto_bandwidth():
    X = RNG.normal(0.0, 1.0, size=(50,))
    res = _gd_gaussian_loc_exact_estimation(
        X, par_v=0.0, par_c=1.0, burn_in=5, n_step=10, bandwidth="auto"
    )
    assert res["bandwidth"] > 0


# --- _sgd_estimation (uses Gaussian for array-valued par_v) ---


def test_sgd_estimation_result_keys():
    X = RNG.normal(2.0, 1.0, size=(30,))
    par_v = np.array([0.0, 1.0])
    model = Gaussian(par_v=par_v.copy(), random_state=0)
    res = _sgd_estimation(
        X,
        par_v.copy(),
        None,
        model,
        kernel="Gaussian",
        burn_in=5,
        n_step=10,
        bandwidth=1.0,
    )
    for key in (
        "par_v_init",
        "par_c_init",
        "stepsize",
        "estimator",
        "trajectory",
        "bandwidth",
    ):
        assert key in res


def test_sgd_estimation_estimator_shape():
    X = RNG.normal(0.0, 1.0, size=(30,))
    par_v = np.array([0.0, 1.0])
    model = Gaussian(par_v=par_v.copy(), random_state=0)
    res = _sgd_estimation(
        X,
        par_v.copy(),
        None,
        model,
        kernel="Gaussian",
        burn_in=5,
        n_step=10,
        bandwidth=1.0,
    )
    assert res["estimator"].shape == (2,)


def test_sgd_estimation_trajectory_columns():
    X = RNG.normal(0.0, 1.0, size=(30,))
    par_v = np.array([0.0, 1.0])
    burn_in, n_step = 5, 10
    model = Gaussian(par_v=par_v.copy(), random_state=0)
    res = _sgd_estimation(
        X,
        par_v.copy(),
        None,
        model,
        kernel="Gaussian",
        burn_in=burn_in,
        n_step=n_step,
        bandwidth=1.0,
    )
    assert res["trajectory"].shape[1] == burn_in + n_step + 1



import numpy as np
from numpy.typing import NDArray
from scipy.special import factorial

from regmmd.models.base_model import RegressionModel
from sklearn.linear_model import PoissonRegressor


class PoissonRegressionBase(RegressionModel):
    def __init__(self, beta=None, random_state=None):
        self.beta = beta

        self.random_state = random_state
        self.rng = np.random.default_rng(seed=self.random_state)

    def log_prob(self, X: NDArray, y: NDArray) -> NDArray:
        if self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the log_prob"
            )
        dot_prod = X @ self.beta

        log_y = y * dot_prod
        log_exp = -np.exp(dot_prod)
        log_Z = np.log(factorial(y))
        return np.sum(log_y + log_exp + log_Z)

    def sample_n(self, n: int, mu_given_x: NDArray) -> NDArray:
        return self.rng.poisson(lam=1 / mu_given_x, size=(n,))

    def predict(self, X: NDArray) -> NDArray:
        """Outputs the mean given X, parameters need to be initialized for this"""
        if self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the predict"
            )

        return np.exp(X @ self.beta)

    def score(self, X: NDArray, y: NDArray) -> NDArray:
        if self.beta is None:
            raise ValueError(
                "Both parameters need to be defined to calculate the score"
            )

        mu = self.predict(X)

        residuals = (y - mu)[:, np.newaxis]
        return residuals * X

    def update(self, par_v):
        self.beta = par_v

    def _project_params(self, par_v):
        return par_v

    def _init_params(self, X, y):
        if self.beta is None:
            init_model = PoissonRegressor(fit_intercept=False).fit(X, y)
            self.beta = init_model.coef_
        return self._get_params()


class PoissonRegression(PoissonRegressionBase):
    def __init__(self, par_v=None, par_c=None, random_state=None):
        super().__init__(beta=par_v, random_state=random_state)

    def _get_params(self):
        par_v = self.beta
        par_c = None
        return par_v, par_c



import numpy as np
import pytest
from regmmd.kernels import K1d_dist, K1d, Kmd, Kmd_dist

KERNELS = ["Gaussian", "Laplace", "Cauchy"]


def test_K1d_dist_gaussian():
    u = np.array([0, 1, -1])
    result = K1d_dist(u, "Gaussian")
    expected = np.exp(-(u**2))
    assert np.allclose(result, expected)


def test_K1d_dist_laplace():
    u = np.array([0, 1, -1])
    result = K1d_dist(u, "Laplace")
    expected = np.exp(-np.abs(u))
    assert np.allclose(result, expected)


def test_K1d_dist_cauchy():
    u = np.array([0, 1, -1])
    result = K1d_dist(u, "Cauchy")
    expected = 1 / (2 + u**2)
    assert np.allclose(result, expected)


def test_K1d_dist_bandwidth():
    u = np.array([0, 1, -1])
    bandwidth = 2.0
    result = K1d_dist(u, "Gaussian", bandwidth=bandwidth)
    scaled_u = u / bandwidth
    expected = np.exp(-(scaled_u**2))
    assert np.allclose(result, expected)


@pytest.mark.parametrize("kernel", KERNELS)
def test_K1d_correct_dims(kernel):
    x = np.array([0, 1, -1])
    y = np.array([1.4, -0.4, 3.1])
    bandwidth = 2.0
    result = K1d(x=x, y=y, kernel=kernel, bandwidth=bandwidth)
    assert result.shape == (3, 3)


def test_K1d_correct_gaussian():
    x = np.array([0, 1, -1])
    y = np.array([1.4, -0.4, 3.1])
    bandwidth = 1.0
    result = K1d(x=x, y=y, kernel="Gaussian", bandwidth=bandwidth)

    expected = np.zeros(shape=(3, 3))
    for i in range(3):
        for j in range(3):
            expected[i, j] = np.exp(-((x[i] - y[j]) ** 2))
    assert np.allclose(result, expected)


def test_K1d_correct_laplace():
    x = np.array([0, 1, -1])
    y = np.array([1.4, -0.4, 3.1])
    bandwidth = 1.0
    result = K1d(x=x, y=y, kernel="Laplace", bandwidth=bandwidth)

    expected = np.zeros(shape=(3, 3))
    for i in range(3):
        for j in range(3):
            expected[i, j] = np.exp(-abs(x[i] - y[j]))
    assert np.allclose(result, expected)


def test_K1d_correct_cauchy():
    x = np.array([0, 1, -1])
    y = np.array([1.4, -0.4, 3.1])
    bandwidth = 1.0
    result = K1d(x=x, y=y, kernel="Cauchy", bandwidth=bandwidth)

    expected = np.zeros(shape=(3, 3))
    for i in range(3):
        for j in range(3):
            expected[i, j] = 1 / (2 + (x[i] - y[j]) ** 2)
    assert np.allclose(result, expected)


def test_K1d_dist_invalid_kernel():
    u = np.array([0, 1, -1])
    with pytest.raises(ValueError):
        K1d_dist(u, "InvalidKernel")


@pytest.mark.parametrize("kernel", KERNELS)
def test_Kmd_correct_dims(kernel):
    u = np.arange(8).reshape(4, 2)
    result = Kmd_dist(u, kernel=kernel)
    assert result.shape == (4,)


@pytest.mark.parametrize("kernel", KERNELS)
def test_Kmd_dist_correct_dims(kernel):
    x = np.arange(8).reshape(4, 2)
    y = -np.arange(8).reshape(4, 2)
    result = Kmd(x, y, kernel=kernel)
    assert result.shape == (4, 4)


def test_Kmd_dist_correct_output_gaussian():
    x = np.arange(8).reshape(4, 2)
    y = -np.arange(8).reshape(4, 2)
    result = Kmd(x, y, kernel="Gaussian")

    expected = np.zeros(shape=(4, 4))
    for i in range(4):
        for j in range(4):
            expected[i, j] = np.exp(-(np.linalg.norm(x[i, :] - y[j, :]) ** 2))
    assert np.allclose(result, expected)


# def test_Kmd_dist_cauchy():
#     u = np.array([0, 1, -1])
#     result = K1d_dist(u, "Cauchy")
#     expected = 1 / (2 + u**2)
#     assert np.allclose(result, expected)



from regmmd.models.estimation.gaussian import Gaussian, GaussianLoc, GaussianScale
from regmmd.models.estimation.beta import Beta, BetaA, BetaB
from regmmd.models.estimation.binomial import Binomial
from regmmd.models.estimation.gamma import Gamma, GammaShape, GammaRate
from regmmd.models.estimation.poisson import Poisson


__all__ = [
    "Gaussian",
    "GaussianLoc",
    "GaussianScale",
    "Beta",
    "BetaA",
    "BetaB",
    "Binomial",
    "Gamma",
    "GammaShape",
    "GammaRate",
    "Poisson",
]
